from typing import TYPE_CHECKING, Any, cast

import litellm
from litellm.utils import supports_response_schema
from pydantic import BaseModel

from any_agent.config import AgentConfig, AgentFramework

from .any_agent import AnyAgent

try:
    from langchain_core.language_models import LanguageModelLike
    from langgraph.prebuilt import create_react_agent

    from any_agent.vendor.langchain_litellm import ChatLiteLLM, _convert_message_to_dict

    DEFAULT_AGENT_TYPE = create_react_agent
    DEFAULT_MODEL_TYPE = ChatLiteLLM

    langchain_available = True
except ImportError:
    langchain_available = False

if TYPE_CHECKING:
    from langchain_core.language_models import LanguageModelLike
    from langgraph.graph.state import CompiledStateGraph


class LangchainAgent(AnyAgent):
    """LangChain agent implementation that handles both loading and running."""

    def __init__(self, config: AgentConfig):
        super().__init__(config)
        self._agent: CompiledStateGraph[Any] | None = None

    @property
    def framework(self) -> AgentFramework:
        return AgentFramework.LANGCHAIN

    def _get_model(self, agent_config: AgentConfig) -> "LanguageModelLike":
        """Get the model configuration for a LangChain agent."""
        model_type = agent_config.model_type or DEFAULT_MODEL_TYPE
        model_args = agent_config.model_args or {}
        return cast(
            "LanguageModelLike",
            model_type(
                model=agent_config.model_id,
                api_key=agent_config.api_key,
                api_base=agent_config.api_base,
                model_kwargs=model_args,  # type: ignore[arg-type]
            ),
        )

    async def _load_agent(self) -> None:
        """Load the LangChain agent with the given configuration."""
        if not langchain_available:
            msg = "You need to `pip install 'any-agent[langchain]'` to use this agent"
            raise ImportError(msg)

        imported_tools, _ = await self._load_tools(self.config.tools)

        self._tools = imported_tools
        agent_type = self.config.agent_type or DEFAULT_AGENT_TYPE
        agent_args = self.config.agent_args or {}
        self._agent = agent_type(
            name=self.config.name,
            model=self._get_model(self.config),
            tools=imported_tools,
            prompt=self.config.instructions,
            **agent_args,
        )

    async def _run_async(self, prompt: str, **kwargs: Any) -> str | BaseModel:
        if not self._agent:
            error_message = "Agent not loaded. Call load_agent() first."
            raise ValueError(error_message)
        inputs = {"messages": [("user", prompt)]}
        result = await self._agent.ainvoke(inputs, **kwargs)

        if not result.get("messages"):
            msg = "No messages returned from the agent."
            raise ValueError(msg)

        # Post-process for structured output if needed
        # This emulates the langgraph behavior for structured outputs,
        # but since it happens outside of langgraph we can control the model call,
        # because some providers like mistral don't allow for the model to be called without the most recent message being a user message.
        if self.config.output_type:
            # Add a follow-up message to request structured output
            structured_output_message = {
                "role": "user",
                "content": f"Please conform your output to the following schema: {self.config.output_type.model_json_schema()}.",
            }
            completion_params: dict[str, Any] = {}
            if self.config.model_args:
                # only include the temperature and frequency_penalty, not anything related to tools
                completion_params["temperature"] = self.config.model_args.get(
                    "temperature"
                )
                completion_params["frequency_penalty"] = self.config.model_args.get(
                    "frequency_penalty"
                )

            completion_params["model"] = self.config.model_id
            previous_messages = [
                _convert_message_to_dict(m) for m in result["messages"]
            ]
            completion_params["messages"] = [
                *previous_messages,
                structured_output_message,
            ]

            # Use response schema if supported by the model
            if supports_response_schema(model=self.config.model_id):
                completion_params["response_format"] = self.config.output_type

            response = await self.call_model(**completion_params)
            return self.config.output_type.model_validate_json(
                response.choices[0].message["content"]
            )
        return str(result["messages"][-1].content)

    async def call_model(self, **kwargs: Any) -> Any:
        return await litellm.acompletion(**kwargs)

    async def update_output_type_async(
        self, output_type: type[BaseModel] | None
    ) -> None:
        """Update the output type of the agent in-place.

        Args:
            output_type: The new output type to use, or None to remove output type constraint

        """
        self.config.output_type = output_type



from collections.abc import Generator, Sequence
from datetime import timedelta
from typing import Any
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from agno.tools.mcp import MCPTools as AgnoMCPTools

from any_agent.config import AgentFramework, MCPParams, MCPSse, MCPStdio, Tool
from any_agent.tools import _get_mcp_server


@pytest.fixture
def agno_mcp_tools() -> Generator[AgnoMCPTools, None, None]:
    with patch("any_agent.tools.mcp.frameworks.agno.AgnoMCPTools") as mock_mcp_tools:
        yield mock_mcp_tools


@pytest.mark.asyncio
@pytest.mark.usefixtures(
    "enter_context_with_transport_and_session",
)
async def test_agno_mcp_sse_integration(
    mcp_sse_params_with_tools: MCPSse,
    session: Any,
    tools: Sequence[Tool],
    agno_mcp_tools: AgnoMCPTools,
) -> None:
    mcp_server = _get_mcp_server(mcp_sse_params_with_tools, AgentFramework.AGNO)

    await mcp_server._setup_tools()

    session.initialize.assert_called_once()

    agno_mcp_tools.assert_called_once_with(session=session, include_tools=tools)  # type: ignore[attr-defined]


@pytest.mark.asyncio
@pytest.mark.usefixtures(
    "enter_context_with_transport_and_session",
)
async def test_agno_mcp_no_tools(
    mcp_params_no_tools: MCPParams,
    agno_mcp_tools: AgnoMCPTools,
) -> None:
    """Regression test:"""
    mcp_server = _get_mcp_server(mcp_params_no_tools, AgentFramework.AGNO)

    await mcp_server._setup_tools()

    assert agno_mcp_tools.call_args_list[0].kwargs["include_tools"] is None  # type: ignore[attr-defined]


@pytest.mark.asyncio
async def test_agno_mcp_env() -> None:
    mcp_server = _get_mcp_server(
        MCPStdio(command="print('Hello MCP')", args=[], env={"FOO": "BAR"}),
        AgentFramework.AGNO,
    )
    mocked_class = MagicMock()
    mocked_cm = AsyncMock()
    mocked_cm.__aenter__.return_value = "foo"
    mocked_class.return_value = mocked_cm

    with patch("any_agent.tools.mcp.frameworks.agno.AgnoMCPTools", mocked_class):
        await mcp_server._setup_tools()
        assert mocked_class.call_args_list[0][1]["env"] == {"FOO": "BAR"}


@pytest.mark.asyncio
async def test_agno_client_session_timeout_passed() -> None:
    """Test that client_session_timeout_seconds parameter is properly passed to AgnoMCPTools (STDIO only)."""
    custom_timeout = 15
    stdio_params = MCPStdio(
        command="echo",
        args=["test"],
        client_session_timeout_seconds=custom_timeout,
    )
    sse_params = MCPSse(
        url="http://localhost:8000",
        client_session_timeout_seconds=custom_timeout,
    )
    # STDIO
    server = _get_mcp_server(stdio_params, AgentFramework.AGNO)
    with patch("any_agent.tools.mcp.frameworks.agno.AgnoMCPTools") as mock_agno_tools:
        mock_tools_instance = AsyncMock()
        mock_tools_instance.__aenter__ = AsyncMock(return_value=mock_tools_instance)
        mock_tools_instance.__aexit__ = AsyncMock(return_value=None)
        mock_agno_tools.return_value = mock_tools_instance
        await server._setup_tools()
        mock_agno_tools.assert_called_once()
        call_args = mock_agno_tools.call_args
        assert call_args.kwargs["timeout_seconds"] == custom_timeout
    # SSE (check that timeout is passed to ClientSession)
    server = _get_mcp_server(sse_params, AgentFramework.AGNO)
    with (
        patch("any_agent.tools.mcp.frameworks.agno.sse_client") as sse_client_patch,
        patch("any_agent.tools.mcp.frameworks.agno.ClientSession") as mock_session,
        patch("any_agent.tools.mcp.frameworks.agno.AgnoMCPTools") as mock_agno_tools,
    ):
        mock_sse_client = AsyncMock()
        mock_sse_client.__aenter__ = AsyncMock(return_value=(MagicMock(), MagicMock()))
        mock_sse_client.__aexit__ = AsyncMock(return_value=None)
        sse_client_patch.return_value = mock_sse_client

        mock_session_instance = AsyncMock()
        mock_session_instance.__aenter__ = AsyncMock(return_value=mock_session_instance)
        mock_session_instance.__aexit__ = AsyncMock(return_value=None)
        mock_session_instance.initialize = AsyncMock()
        mock_session.return_value = mock_session_instance

        mock_tools_instance = AsyncMock()
        mock_tools_instance.__aenter__ = AsyncMock(return_value=mock_tools_instance)
        mock_tools_instance.__aexit__ = AsyncMock(return_value=None)
        mock_agno_tools.return_value = mock_tools_instance

        await server._setup_tools()
        mock_session.assert_called_once()
        # Check that the timeout was passed as read_timeout_seconds
        call_args = mock_session.call_args
        assert call_args.kwargs["read_timeout_seconds"] == timedelta(
            seconds=custom_timeout
        )



# mypy: disable-error-code="no-untyped-def,override,union-attr"
from __future__ import annotations

from typing import TYPE_CHECKING, Any

from any_agent.callbacks.span_generation.base import _SpanGeneration

if TYPE_CHECKING:
    from google.adk.models.llm_request import LlmRequest
    from google.adk.models.llm_response import LlmResponse
    from google.adk.tools.base_tool import BaseTool
    from google.adk.tools.tool_context import ToolContext

    from any_agent.callbacks.context import Context


class _GoogleSpanGeneration(_SpanGeneration):
    def before_llm_call(self, context: Context, *args: Any, **kwargs: Any) -> Context:
        llm_request: LlmRequest = kwargs["llm_request"]

        messages = []
        if config := llm_request.config:
            messages.append(
                {
                    "role": "system",
                    "content": getattr(config, "system_instruction", "No instructions"),
                }
            )
        if parts := llm_request.contents[0].parts:
            messages.append(
                {
                    "role": getattr(llm_request.contents[0], "role", "No role"),
                    "content": getattr(parts[0], "text", "No content"),
                }
            )

        return self._set_llm_input(context, str(llm_request.model), messages)

    def after_llm_call(self, context: Context, *args, **kwargs) -> Context:
        llm_response: LlmResponse = kwargs["llm_response"]

        content = llm_response.content
        output: str | list[dict[str, Any]]
        if not content or not content.parts:
            output = ""
        elif content.parts[0].text:
            output = str(content.parts[0].text)
        else:
            output = [
                {
                    "tool.name": getattr(part.function_call, "name", "No name"),
                    "tool.args": getattr(part.function_call, "args", "{}"),
                }
                for part in content.parts
                if part.function_call
            ]
        input_tokens = 0
        output_tokens = 0
        if resp_meta := llm_response.usage_metadata:
            if prompt_tokens := resp_meta.prompt_token_count:
                input_tokens = prompt_tokens
            if candidates_token := resp_meta.candidates_token_count:
                output_tokens = candidates_token
        return self._set_llm_output(context, output, input_tokens, output_tokens)

    def before_tool_execution(self, context: Context, *args, **kwargs) -> Context:
        tool: BaseTool = kwargs["tool"]
        tool_args: dict[str, Any] = kwargs["args"]
        tool_context: ToolContext = kwargs["tool_context"]

        return self._set_tool_input(
            context,
            tool.name,
            tool.description,
            tool_args,
            tool_context.function_call_id,
        )

    def after_tool_execution(self, context: Context, *args, **kwargs) -> Context:
        return self._set_tool_output(context, kwargs["tool_response"])



from unittest.mock import MagicMock

from any_agent import AgentFramework
from any_agent.callbacks.wrappers import _get_wrapper_by_framework


async def test_unwrap_before_wrap(agent_framework: AgentFramework) -> None:
    wrapper = _get_wrapper_by_framework(agent_framework)
    await wrapper.unwrap(MagicMock())


async def test_google_instrument_uninstrument() -> None:
    """Regression test for https://github.com/mozilla-ai/any-agent/issues/467"""
    agent = MagicMock()
    agent._agent.before_model_callback = None
    agent._agent.after_model_callback = None
    agent._agent.before_tool_callback = None
    agent._agent.after_tool_callback = None

    wrapper = _get_wrapper_by_framework(AgentFramework.GOOGLE)

    await wrapper.wrap(agent)
    assert callable(agent._agent.before_model_callback)
    assert callable(agent._agent.after_model_callback)
    assert callable(agent._agent.before_tool_callback)
    assert callable(agent._agent.after_tool_callback)

    await wrapper.unwrap(agent)
    assert agent._agent.before_model_callback is None
    assert agent._agent.after_model_callback is None
    assert agent._agent.before_tool_callback is None
    assert agent._agent.after_tool_callback is None



from unittest.mock import MagicMock

from any_agent.callbacks.span_cost import AddCostInfo
from any_agent.testing.helpers import DEFAULT_SMALL_MODEL_ID
from any_agent.tracing.attributes import GenAI


def test_span_cost() -> None:
    context = MagicMock()
    current_span = MagicMock()

    current_span.attributes = {
        GenAI.REQUEST_MODEL: DEFAULT_SMALL_MODEL_ID,
        GenAI.USAGE_INPUT_TOKENS: 100,
        GenAI.USAGE_OUTPUT_TOKENS: 1000,
    }

    context.current_span = current_span

    callback = AddCostInfo()

    callback.after_llm_call(context)

    call_args = context.current_span.set_attributes.call_args[0][0]
    assert call_args[GenAI.USAGE_INPUT_COST] > 0
    assert call_args[GenAI.USAGE_OUTPUT_COST] > 0


def test_span_cost_missing_input() -> None:
    context = MagicMock()
    current_span = MagicMock()

    current_span.attributes = {
        GenAI.REQUEST_MODEL: DEFAULT_SMALL_MODEL_ID,
        GenAI.USAGE_OUTPUT_TOKENS: 1000,
    }

    context.current_span = current_span

    callback = AddCostInfo()

    callback.after_llm_call(context)

    call_args = context.current_span.set_attributes.call_args[0][0]
    assert call_args[GenAI.USAGE_INPUT_COST] == 0
    assert call_args[GenAI.USAGE_OUTPUT_COST] > 0


def test_span_cost_missing_output() -> None:
    context = MagicMock()
    current_span = MagicMock()

    current_span.attributes = {
        GenAI.REQUEST_MODEL: DEFAULT_SMALL_MODEL_ID,
        GenAI.USAGE_INPUT_TOKENS: 100,
    }

    context.current_span = current_span

    callback = AddCostInfo()

    callback.after_llm_call(context)

    call_args = context.current_span.set_attributes.call_args[0][0]
    assert call_args[GenAI.USAGE_INPUT_COST] > 0
    assert call_args[GenAI.USAGE_OUTPUT_COST] == 0


def test_span_cost_missing_all() -> None:
    context = MagicMock()
    current_span = MagicMock()

    current_span.attributes = {
        GenAI.REQUEST_MODEL: DEFAULT_SMALL_MODEL_ID,
    }

    context.current_span = current_span

    callback = AddCostInfo()

    callback.after_llm_call(context)

    context.current_span.set_attributes.assert_not_called()



import os
import pathlib
import subprocess

import pytest


@pytest.mark.parametrize(
    "notebook_path",
    list(pathlib.Path("docs/cookbook").glob("*.ipynb")),
    ids=lambda x: x.stem,
)
@pytest.mark.timeout(180)  # 3+ minutes timeout for cookbook execution
def test_cookbook_notebook(
    notebook_path: pathlib.Path, capsys: pytest.CaptureFixture
) -> None:
    """Test that cookbook notebooks execute without errors using jupyter execute."""
    try:
        result = subprocess.run(  # noqa: S603
            ["jupyter", "execute", notebook_path.name],  # noqa: S607
            cwd="docs/cookbook",  # Run in cookbook directory like original action
            env={
                "MISTRAL_API_KEY": os.environ["MISTRAL_API_KEY"],
                "TAVILY_API_KEY": os.environ["TAVILY_API_KEY"],
                "OPENAI_API_KEY": os.environ["OPENAI_API_KEY"],
                "PATH": os.environ["PATH"],
                "IN_PYTEST": "1",  # For mcp_agent notebook which needs to mock input
            },
            timeout=170,  # Time out slightly earlier so that we can log the output.
            capture_output=True,
            check=False,
        )
    except subprocess.TimeoutExpired as e:
        # Handle timeout case - log stdout/stderr that were captured before timeout
        stdout = e.stdout.decode() if e.stdout else "(no stdout captured)"
        stderr = e.stderr.decode() if e.stderr else "(no stderr captured)"
        pytest.fail(
            f"Notebook {notebook_path.name} timed out after 2 minutes\n stdout: {stdout}\n stderr: {stderr}"
        )

    if result.returncode != 0:
        stdout = result.stdout.decode() if result.stdout else "(no stdout captured)"
        stderr = result.stderr.decode() if result.stderr else "(no stderr captured)"
        pytest.fail(
            f"Notebook {notebook_path.name} failed with return code {result.returncode}\n stdout: {stdout}\n stderr: {stderr}"
        )



from .frameworks import (
    MCPServer,
    _get_mcp_server,
)
from .mcp_connection import _MCPConnection
from .mcp_server import _MCPServerBase

__all__ = [
    "MCPServer",
    "_MCPConnection",
    "_MCPServerBase",
    "_get_mcp_server",
]



import logging

from any_agent.logging import setup_logger


def test_setup_logger_sets_level_and_propagate(local_logger: logging.Logger) -> None:
    setup_logger(level=logging.INFO, propagate=False)
    assert local_logger.level == logging.INFO
    assert local_logger.propagate is False
    setup_logger(level=logging.DEBUG, propagate=True)
    assert local_logger.level == logging.DEBUG
    assert local_logger.propagate is True


def test_setup_logger_removes_existing_handlers(local_logger: logging.Logger) -> None:
    dummy_handler = logging.StreamHandler()
    local_logger.addHandler(dummy_handler)
    assert dummy_handler in local_logger.handlers
    setup_logger()
    assert dummy_handler not in local_logger.handlers


def test_setup_logger_with_custom_format(local_logger: logging.Logger) -> None:
    custom_format = "%(levelname)s: %(message)s"
    setup_logger(log_format=custom_format)
    handler = local_logger.handlers[0]
    assert isinstance(handler.formatter, logging.Formatter)
    assert handler.formatter._fmt == custom_format


def test_setup_logger_multiple_calls_idempotent(local_logger: logging.Logger) -> None:
    setup_logger(level=logging.INFO)
    first_handler = local_logger.handlers[0]
    setup_logger(level=logging.WARNING)
    second_handler = local_logger.handlers[0]
    assert first_handler is not second_handler
    assert local_logger.level == logging.WARNING


def test_setup_logger_edits_global_logger(local_logger: logging.Logger) -> None:
    # This test now just checks that setup_logger edits the patched logger
    dummy_handler = logging.StreamHandler()
    local_logger.addHandler(dummy_handler)
    assert dummy_handler in local_logger.handlers
    setup_logger()
    assert dummy_handler not in local_logger.handlers
    assert local_logger.level == logging.ERROR
    assert local_logger.propagate is False



from abc import ABC, abstractmethod
from collections.abc import Callable, Sequence
from contextlib import suppress
from datetime import timedelta
from typing import Any, Literal, Union

from pydantic import Field, PrivateAttr

from any_agent.config import AgentFramework, MCPSse, MCPStdio, MCPStreamableHttp
from any_agent.tools.mcp.mcp_connection import _MCPConnection
from any_agent.tools.mcp.mcp_server import _MCPServerBase

mcp_available = False
with suppress(ImportError):
    from langchain_core.tools import BaseTool  # noqa: TC002
    from langchain_mcp_adapters.tools import load_mcp_tools
    from mcp import ClientSession, StdioServerParameters
    from mcp.client.sse import sse_client
    from mcp.client.stdio import stdio_client
    from mcp.client.streamable_http import streamablehttp_client

    mcp_available = True


class LangchainMCPConnection(_MCPConnection["BaseTool"], ABC):
    """Base class for LangChain MCP connections."""

    _client: Any | None = PrivateAttr(default=None)
    session: Union["ClientSession", None] = None

    @abstractmethod
    async def list_tools(self) -> list["BaseTool"]:
        """List tools from the MCP server."""
        if not self._client:
            msg = "MCP client is not set up. Please call `list_tools` from a concrete class."
            raise ValueError(msg)

        tools = await load_mcp_tools(self.session)
        return self._filter_tools(tools)  # type: ignore[return-value]


class LangchainMCPStdioConnection(LangchainMCPConnection):
    mcp_tool: MCPStdio

    async def list_tools(self) -> list["BaseTool"]:
        """List tools from the MCP server."""
        server_params = StdioServerParameters(
            command=self.mcp_tool.command,
            args=list(self.mcp_tool.args),
            env=self.mcp_tool.env,
        )

        self._client = stdio_client(server_params)

        read, write = await self._exit_stack.enter_async_context(self._client)
        kwargs = {}
        if self.mcp_tool.client_session_timeout_seconds:
            kwargs["read_timeout_seconds"] = timedelta(
                seconds=self.mcp_tool.client_session_timeout_seconds
            )
        client_session = ClientSession(
            read,
            write,
            **kwargs,  # type: ignore[arg-type]
        )
        self.session = await self._exit_stack.enter_async_context(client_session)

        await self.session.initialize()

        return await super().list_tools()


class LangchainMCPSseConnection(LangchainMCPConnection):
    mcp_tool: MCPSse

    async def list_tools(self) -> list["BaseTool"]:
        """List tools from the MCP server."""
        sse_kwargs = {}
        if self.mcp_tool.client_session_timeout_seconds:
            sse_kwargs["sse_read_timeout"] = (
                self.mcp_tool.client_session_timeout_seconds
            )
        self._client = sse_client(
            url=self.mcp_tool.url,
            headers=dict(self.mcp_tool.headers or {}),
            **sse_kwargs,  # type: ignore[arg-type]
        )
        read, write = await self._exit_stack.enter_async_context(self._client)
        session_kwargs = {}
        if self.mcp_tool.client_session_timeout_seconds:
            session_kwargs["read_timeout_seconds"] = timedelta(
                seconds=self.mcp_tool.client_session_timeout_seconds
            )
        client_session = ClientSession(
            read,
            write,
            **session_kwargs,  # type: ignore[arg-type]
        )
        self.session = await self._exit_stack.enter_async_context(client_session)

        await self.session.initialize()

        return await super().list_tools()


class LangchainMCPStreamableHttpConnection(LangchainMCPConnection):
    mcp_tool: MCPStreamableHttp
    _get_session_id_callback: Callable[[], str | None] | None = None

    def get_session_id(self) -> str | None:
        """Retrieve session ID, if it has been established."""
        if self._get_session_id_callback:
            return self._get_session_id_callback()
        return None

    async def list_tools(self) -> list["BaseTool"]:
        """List tools from the MCP server."""
        http_kwargs = {}
        if self.mcp_tool.client_session_timeout_seconds:
            http_kwargs["sse_read_timeout"] = timedelta(
                seconds=self.mcp_tool.client_session_timeout_seconds
            )
        self._client = streamablehttp_client(
            url=self.mcp_tool.url,
            headers=dict(self.mcp_tool.headers or {}),
            **http_kwargs,  # type: ignore[arg-type]
        )
        streamablehttp_transport = await self._exit_stack.enter_async_context(
            self._client
        )
        read_stream, write_stream, get_session_id_callback = streamablehttp_transport
        self._get_session_id_callback = get_session_id_callback
        session_kwargs = {}
        if self.mcp_tool.client_session_timeout_seconds:
            session_kwargs["read_timeout_seconds"] = timedelta(
                seconds=self.mcp_tool.client_session_timeout_seconds
            )
        client_session = ClientSession(read_stream, write_stream, **session_kwargs)  # type: ignore[arg-type]
        self.session = await self._exit_stack.enter_async_context(client_session)
        await self.session.initialize()
        return await super().list_tools()


class LangchainMCPServerBase(_MCPServerBase["BaseTool"], ABC):
    framework: Literal[AgentFramework.LANGCHAIN] = AgentFramework.LANGCHAIN
    tools: Sequence["BaseTool"] = Field(default_factory=list)

    def _check_dependencies(self) -> None:
        self.libraries = "any-agent[mcp,langchain]"
        self.mcp_available = mcp_available
        super()._check_dependencies()


class LangchainMCPServerStdio(LangchainMCPServerBase):
    mcp_tool: MCPStdio

    async def _setup_tools(
        self, mcp_connection: _MCPConnection["BaseTool"] | None = None
    ) -> None:
        mcp_connection = mcp_connection or LangchainMCPStdioConnection(
            mcp_tool=self.mcp_tool
        )
        await super()._setup_tools(mcp_connection)


class LangchainMCPServerSse(LangchainMCPServerBase):
    mcp_tool: MCPSse

    async def _setup_tools(
        self, mcp_connection: _MCPConnection["BaseTool"] | None = None
    ) -> None:
        mcp_connection = mcp_connection or LangchainMCPSseConnection(
            mcp_tool=self.mcp_tool
        )
        await super()._setup_tools(mcp_connection)


class LangchainMCPServerStreamableHttp(LangchainMCPServerBase):
    mcp_tool: MCPStreamableHttp

    async def _setup_tools(
        self, mcp_connection: _MCPConnection["BaseTool"] | None = None
    ) -> None:
        mcp_connection = mcp_connection or LangchainMCPStreamableHttpConnection(
            mcp_tool=self.mcp_tool
        )
        await super()._setup_tools(mcp_connection)


LangchainMCPServer = (
    LangchainMCPServerStdio | LangchainMCPServerSse | LangchainMCPServerStreamableHttp
)



from typing import TYPE_CHECKING, Any

from pydantic import BaseModel

from any_agent.config import AgentConfig, AgentFramework

from .any_agent import AnyAgent

try:
    from agents import (
        Agent,
        Model,
        ModelSettings,
        Runner,
    )
    from agents.extensions.models.litellm_model import LitellmModel

    DEFAULT_MODEL_TYPE = LitellmModel

    agents_available = True
except ImportError:
    agents_available = False


if TYPE_CHECKING:
    from agents import Model


class OpenAIAgent(AnyAgent):
    """OpenAI agent implementation that handles both loading and running."""

    def __init__(self, config: AgentConfig):
        super().__init__(config)
        self._agent: Agent | None = None

    @property
    def framework(self) -> AgentFramework:
        return AgentFramework.OPENAI

    def _get_model(
        self,
        agent_config: AgentConfig,
    ) -> "Model":
        """Get the model configuration for an OpenAI agent."""
        model_type = agent_config.model_type or DEFAULT_MODEL_TYPE
        return model_type(
            model=agent_config.model_id,
            base_url=agent_config.api_base,
            api_key=agent_config.api_key,
        )

    async def _load_agent(self) -> None:
        """Load the OpenAI agent with the given configuration."""
        if not agents_available:
            msg = "You need to `pip install 'any-agent[openai]'` to use this agent"
            raise ImportError(msg)
        if not agents_available:
            msg = "You need to `pip install openai-agents` to use this agent"
            raise ImportError(msg)

        tools, mcp_servers = await self._load_tools(self.config.tools)
        tools = self._filter_mcp_tools(tools, mcp_servers)

        kwargs_ = self.config.agent_args or {}
        if self.config.model_args:
            kwargs_["model_settings"] = ModelSettings(**self.config.model_args)
        if self.config.output_type:
            kwargs_["output_type"] = self.config.output_type

        self._tools = tools
        self._agent = Agent(
            name=self.config.name,
            instructions=self.config.instructions,
            model=self._get_model(self.config),
            tools=tools,
            mcp_servers=[mcp_server.server for mcp_server in mcp_servers],
            **kwargs_,
        )

    def _filter_mcp_tools(self, tools: list[Any], mcp_servers: list[Any]) -> list[Any]:
        """OpenAI framework doesn't expect the mcp tool to be included in `tools`."""
        non_mcp_tools = []
        for tool in tools:
            if any(tool in mcp_server.tools for mcp_server in mcp_servers):
                continue
            non_mcp_tools.append(tool)
        return non_mcp_tools

    async def _run_async(self, prompt: str, **kwargs: Any) -> str | BaseModel:
        if not self._agent:
            error_message = "Agent not loaded. Call load_agent() first."
            raise ValueError(error_message)
        result = await Runner.run(self._agent, prompt, **kwargs)
        return result.final_output  # type: ignore[no-any-return]

    async def update_output_type_async(
        self, output_type: type[BaseModel] | None
    ) -> None:
        """Update the output type of the agent in-place.

        Args:
            output_type: The new output type to use, or None to remove output type constraint

        """
        self.config.output_type = output_type

        # If agent is already loaded, we need to recreate it with the new output type
        # The OpenAI agents library requires output_type to be set during construction
        if self._agent:
            # Store current state
            current_tools = self._tools
            current_mcp_servers = self._mcp_servers

            # Recreate the agent with the new output type
            kwargs_ = self.config.agent_args or {}
            if self.config.model_args:
                kwargs_["model_settings"] = ModelSettings(**self.config.model_args)
            if output_type:
                kwargs_["output_type"] = output_type

            self._agent = Agent(
                name=self.config.name,
                instructions=self.config.instructions,
                model=self._get_model(self.config),
                tools=current_tools,
                mcp_servers=[mcp_server.server for mcp_server in current_mcp_servers],
                **kwargs_,
            )



from collections.abc import Callable
from typing import TYPE_CHECKING, Any

from pydantic import BaseModel

from any_agent.config import AgentConfig, AgentFramework
from any_agent.frameworks.any_agent import AnyAgent
from any_agent.tools.final_output import prepare_final_output

try:
    from smolagents import FinalAnswerTool, LiteLLMModel, ToolCallingAgent

    smolagents_available = True
except ImportError:
    smolagents_available = False

if TYPE_CHECKING:
    from smolagents import MultiStepAgent


DEFAULT_AGENT_TYPE = ToolCallingAgent
DEFAULT_MODEL_TYPE = LiteLLMModel


class SmolagentsAgent(AnyAgent):
    """Smolagents agent implementation that handles both loading and running."""

    def __init__(
        self,
        config: AgentConfig,
    ):
        super().__init__(config)
        self._agent: MultiStepAgent | None = None

    @property
    def framework(self) -> AgentFramework:
        return AgentFramework.SMOLAGENTS

    def _get_model(self, agent_config: AgentConfig) -> Any:
        """Get the model configuration for a smolagents agent."""
        model_type = agent_config.model_type or DEFAULT_MODEL_TYPE
        model_args = agent_config.model_args or {}
        kwargs = {
            "model_id": agent_config.model_id,
            "api_key": agent_config.api_key,
            "api_base": agent_config.api_base,
            **model_args,
        }
        return model_type(**kwargs)

    def _setup_output_type(self, output_type: type[BaseModel] | None) -> None:
        """Set up the output type handling for the agent.

        Args:
            output_type: The output type to set up, or None to remove output type constraint

        """
        if not self._agent:
            return

        if "final_answer" in self._agent.tools:
            del self._agent.tools["final_answer"]

        if self.config.instructions:
            self._agent.prompt_templates["system_prompt"] = self.config.instructions

        if output_type:
            instructions, final_output_function = prepare_final_output(
                output_type, self.config.instructions
            )

            class FinalAnswerToolWrapper(FinalAnswerTool):  # type: ignore[no-untyped-call]
                def __init__(
                    self,
                    final_output_func: Callable[
                        [str], dict[str, str | bool | dict[str, Any] | list[Any]]
                    ],
                ):
                    super().__init__()  # type: ignore[no-untyped-call]
                    self.final_output_func = final_output_func
                    # Copying the __doc__ relies upon the final_output_func having a single str parameter called "answer"
                    if (
                        not self.final_output_func.__code__.co_varnames[0] == "answer"
                        or not self.final_output_func.__doc__
                    ):
                        msg = "The final_output_func must have a single parameter of type str"
                        raise ValueError(msg)

                    self.inputs = {
                        "answer": {
                            "type": "string",
                            "description": self.final_output_func.__doc__,
                        }
                    }

                def forward(self, answer: str) -> Any:
                    result = self.final_output_func(answer)
                    if result.get("success"):
                        return answer
                    raise ValueError(result["result"])

            self._agent.tools["final_answer"] = FinalAnswerToolWrapper(
                final_output_function
            )

            # Update the system prompt with the modified instructions
            if instructions:
                self._agent.prompt_templates["system_prompt"] = instructions

    async def _load_agent(self) -> None:
        """Load the Smolagents agent with the given configuration."""
        if not smolagents_available:
            msg = "You need to `pip install 'any-agent[smolagents]'` to use this agent"
            raise ImportError(msg)

        tools, _ = await self._load_tools(self.config.tools)

        main_agent_type = self.config.agent_type or DEFAULT_AGENT_TYPE

        agent_args = self.config.agent_args or {}

        self._tools = tools
        self._agent = main_agent_type(
            name=self.config.name,
            model=self._get_model(self.config),
            tools=tools,
            verbosity_level=-1,  # OFF
            **agent_args,
        )

        if self.config.instructions:
            self._agent.prompt_templates["system_prompt"] = self.config.instructions

        # Set up output type handling
        self._setup_output_type(self.config.output_type)

        assert self._agent

    async def _run_async(self, prompt: str, **kwargs: Any) -> str | BaseModel:
        if not self._agent:
            error_message = "Agent not loaded. Call load_agent() first."
            raise ValueError(error_message)
        result = self._agent.run(prompt, **kwargs)
        if self.config.output_type:
            return self.config.output_type.model_validate_json(result)
        return str(result)

    async def update_output_type_async(
        self, output_type: type[BaseModel] | None
    ) -> None:
        """Update the output type of the agent in-place.

        Args:
            output_type: The new output type to use, or None to remove output type constraint

        """
        self.config.output_type = output_type

        # If agent is already loaded, update its output handling
        self._setup_output_type(output_type)



# mypy: disable-error-code="arg-type,attr-defined,no-untyped-def,union-attr"
from __future__ import annotations

import json
from typing import TYPE_CHECKING

from rich.console import Console, Group
from rich.json import JSON
from rich.markdown import Markdown
from rich.panel import Panel

from any_agent.callbacks.base import Callback
from any_agent.tracing.attributes import GenAI

if TYPE_CHECKING:
    from opentelemetry.sdk.trace import ReadableSpan

    from any_agent.callbacks.context import Context


def _get_output_panel(span: ReadableSpan) -> Panel | None:
    if output := span.attributes.get(GenAI.OUTPUT, None):
        output_type = span.attributes.get(GenAI.OUTPUT_TYPE, "text")
        return Panel(
            Markdown(output) if output_type != "json" else JSON(output),
            title="OUTPUT",
            style="white",
            title_align="left",
        )
    return None


class ConsolePrintSpan(Callback):
    """Use rich's console to print the `Context.current_span`."""

    def __init__(self, console: Console | None = None) -> None:
        """Init the ConsolePrintSpan.

        Args:
            console: An optional instance of `rich.console.Console`.
                If `None`, a new instance will be used.

        """
        self.console = console or Console()

    def after_llm_call(self, context: Context, *args, **kwargs) -> Context:
        span = context.current_span

        operation_name = span.attributes.get(GenAI.OPERATION_NAME, "")

        if operation_name != "call_llm":
            return context

        panels = []

        if messages := span.attributes.get(GenAI.INPUT_MESSAGES):
            panels.append(
                Panel(JSON(messages), title="INPUT", style="white", title_align="left")
            )

        if output_panel := _get_output_panel(span):
            panels.append(output_panel)

        if usage := {
            k.replace("gen_ai.usage.", ""): v
            for k, v in span.attributes.items()
            if "usage" in k
        }:
            panels.append(
                Panel(
                    JSON(json.dumps(usage)),
                    title="USAGE",
                    style="white",
                    title_align="left",
                )
            )

        self.console.print(
            Panel(
                Group(*panels),
                title=f"{operation_name.upper()}: {span.attributes.get(GenAI.REQUEST_MODEL)}",
                style="yellow",
            )
        )

        return context

    def after_tool_execution(self, context: Context, *args, **kwargs) -> Context:
        span = context.current_span

        operation_name = span.attributes.get(GenAI.OPERATION_NAME, "")

        if operation_name != "execute_tool":
            return context

        panels = [
            Panel(
                JSON(span.attributes.get(GenAI.TOOL_ARGS, "{}")),
                title="Input",
                style="white",
                title_align="left",
            )
        ]

        if output_panel := _get_output_panel(span):
            panels.append(output_panel)

        self.console.print(
            Panel(
                Group(*panels),
                title=f"{operation_name.upper()}: {span.attributes.get(GenAI.TOOL_NAME)}",
                style="blue",
            )
        )
        return context



from typing import TYPE_CHECKING
from unittest.mock import ANY, AsyncMock, MagicMock, patch

import pytest
from langchain_core.messages import AIMessage
from pydantic import BaseModel

from any_agent import AgentConfig, AgentFramework, AnyAgent
from any_agent.testing.helpers import LITELLM_IMPORT_PATHS

if TYPE_CHECKING:
    from any_agent.frameworks.langchain import LangchainAgent


def test_load_langchain_agent_default() -> None:
    model_mock = MagicMock()
    create_mock = MagicMock()
    agent_mock = MagicMock()
    create_mock.return_value = agent_mock
    tool_mock = MagicMock()

    with (
        patch("any_agent.frameworks.langchain.DEFAULT_AGENT_TYPE", create_mock),
        patch("any_agent.frameworks.langchain.DEFAULT_MODEL_TYPE", model_mock),
        patch("langchain_core.tools.tool", tool_mock),
    ):
        AnyAgent.create(
            AgentFramework.LANGCHAIN,
            AgentConfig(model_id="mistral/mistral-small-latest"),
        )

        model_mock.assert_called_once_with(
            model="mistral/mistral-small-latest",
            api_base=None,
            api_key=None,
            model_kwargs={},
        )
        create_mock.assert_called_once_with(
            name="any_agent",
            model=model_mock.return_value,
            tools=[],
            prompt=None,
        )


def test_load_langchain_agent_missing() -> None:
    with patch("any_agent.frameworks.langchain.langchain_available", False):
        with pytest.raises(ImportError):
            AnyAgent.create(
                AgentFramework.LANGCHAIN,
                AgentConfig(model_id="mistral/mistral-small-latest"),
            )


def test_run_langchain_agent_custom_args() -> None:
    create_mock = MagicMock()
    agent_mock = AsyncMock()
    agent_mock.ainvoke.return_value = MagicMock()
    create_mock.return_value = agent_mock

    with (
        patch("any_agent.frameworks.langchain.DEFAULT_AGENT_TYPE", create_mock),
        patch("any_agent.frameworks.langchain.DEFAULT_MODEL_TYPE"),
        patch("langchain_core.tools.tool"),
    ):
        agent = AnyAgent.create(
            AgentFramework.LANGCHAIN,
            AgentConfig(model_id="mistral/mistral-small-latest"),
        )
        agent.run("foo", debug=True)
        agent_mock.ainvoke.assert_called_once_with(
            {"messages": [("user", "foo")]}, debug=True, config={"callbacks": [ANY]}
        )


class SampleOutput(BaseModel):
    answer: str
    confidence: float


def test_structured_output_without_tools() -> None:
    """Test that structured output works correctly when no tools are present and tool-related params are not set."""
    config = AgentConfig(model_id="gpt-4.1-mini", output_type=SampleOutput)
    agent: LangchainAgent = AnyAgent.create(AgentFramework.LANGCHAIN, config)  # type: ignore[assignment]

    # Patch the agent's _agent to return a mock result for ainvoke
    mock_agent = AsyncMock()
    mock_agent.ainvoke.return_value = {
        "messages": [AIMessage(content="Initial response")]
    }
    agent._agent = mock_agent

    def create_mock_response(content: str) -> MagicMock:
        mock_message = MagicMock()
        mock_message.content = content
        mock_message.__getitem__.side_effect = (
            lambda key: content if key == "content" else None
        )
        return MagicMock(choices=[MagicMock(message=mock_message)])

    with patch(LITELLM_IMPORT_PATHS[AgentFramework.LANGCHAIN]) as mock_acompletion:
        mock_acompletion.return_value = create_mock_response(
            '{"answer": "Structured answer", "confidence": 0.95}'
        )
        agent.run("Test question")

        # Only expect that acompletion was called once for structured output
        assert mock_acompletion.call_count == 1
        call_args = mock_acompletion.call_args[1]
        # Should not include any tool-related keys
        assert "tools" not in call_args
        assert "tool_choice" not in call_args
        # Should include response_format
        assert "response_format" in call_args
        assert call_args["response_format"] == SampleOutput



from typing import assert_never

from any_agent.config import (
    AgentFramework,
    MCPParams,
    MCPSse,
    MCPStdio,
    MCPStreamableHttp,
)

from .agno import AgnoMCPServer
from .google import GoogleMCPServer
from .langchain import LangchainMCPServer
from .llama_index import LlamaIndexMCPServer
from .openai import OpenAIMCPServer
from .smolagents import SmolagentsMCPServer
from .tinyagent import TinyAgentMCPServer

MCPServer = (
    AgnoMCPServer
    | GoogleMCPServer
    | LangchainMCPServer
    | LlamaIndexMCPServer
    | OpenAIMCPServer
    | SmolagentsMCPServer
    | TinyAgentMCPServer
)


def _get_stdio_mcp_server(
    mcp_tool: MCPStdio, agent_framework: AgentFramework
) -> MCPServer:
    if agent_framework is AgentFramework.AGNO:
        from .agno import AgnoMCPServerStdio

        return AgnoMCPServerStdio(mcp_tool=mcp_tool)
    if agent_framework is AgentFramework.GOOGLE:
        from .google import GoogleMCPServerStdio

        return GoogleMCPServerStdio(mcp_tool=mcp_tool)
    if agent_framework is AgentFramework.LANGCHAIN:
        from .langchain import LangchainMCPServerStdio

        return LangchainMCPServerStdio(mcp_tool=mcp_tool)
    if agent_framework is AgentFramework.LLAMA_INDEX:
        from .llama_index import LlamaIndexMCPServerStdio

        return LlamaIndexMCPServerStdio(mcp_tool=mcp_tool)
    if agent_framework is AgentFramework.OPENAI:
        from .openai import OpenAIMCPServerStdio

        return OpenAIMCPServerStdio(mcp_tool=mcp_tool)
    if agent_framework is AgentFramework.SMOLAGENTS:
        from .smolagents import SmolagentsMCPServerStdio

        return SmolagentsMCPServerStdio(mcp_tool=mcp_tool)
    if agent_framework is AgentFramework.TINYAGENT:
        from .tinyagent import TinyAgentMCPServerStdio

        return TinyAgentMCPServerStdio(mcp_tool=mcp_tool)
    assert_never(agent_framework)


def _get_sse_mcp_server(mcp_tool: MCPSse, agent_framework: AgentFramework) -> MCPServer:
    if agent_framework is AgentFramework.AGNO:
        from .agno import AgnoMCPServerSse

        return AgnoMCPServerSse(mcp_tool=mcp_tool)
    if agent_framework is AgentFramework.GOOGLE:
        from .google import GoogleMCPServerSse

        return GoogleMCPServerSse(mcp_tool=mcp_tool)
    if agent_framework is AgentFramework.LANGCHAIN:
        from .langchain import LangchainMCPServerSse

        return LangchainMCPServerSse(mcp_tool=mcp_tool)
    if agent_framework is AgentFramework.LLAMA_INDEX:
        from .llama_index import LlamaIndexMCPServerSse

        return LlamaIndexMCPServerSse(mcp_tool=mcp_tool)
    if agent_framework is AgentFramework.OPENAI:
        from .openai import OpenAIMCPServerSse

        return OpenAIMCPServerSse(mcp_tool=mcp_tool)
    if agent_framework is AgentFramework.SMOLAGENTS:
        from .smolagents import SmolagentsMCPServerSse

        return SmolagentsMCPServerSse(mcp_tool=mcp_tool)
    if agent_framework is AgentFramework.TINYAGENT:
        from .tinyagent import TinyAgentMCPServerSse

        return TinyAgentMCPServerSse(mcp_tool=mcp_tool)
    assert_never(agent_framework)


def _get_streamablehttp_mcp_server(
    mcp_tool: MCPStreamableHttp, agent_framework: AgentFramework
) -> MCPServer:
    if agent_framework is AgentFramework.AGNO:
        from .agno import AgnoMCPServerStreamableHttp

        return AgnoMCPServerStreamableHttp(mcp_tool=mcp_tool)
    if agent_framework is AgentFramework.GOOGLE:
        from .google import GoogleMCPServerStreamableHttp

        return GoogleMCPServerStreamableHttp(mcp_tool=mcp_tool)
    if agent_framework is AgentFramework.LANGCHAIN:
        from .langchain import LangchainMCPServerStreamableHttp

        return LangchainMCPServerStreamableHttp(mcp_tool=mcp_tool)
    if agent_framework is AgentFramework.LLAMA_INDEX:
        from .llama_index import LlamaIndexMCPServerStreamableHttp

        return LlamaIndexMCPServerStreamableHttp(mcp_tool=mcp_tool)
    if agent_framework is AgentFramework.OPENAI:
        from .openai import OpenAIMCPServerStreamableHttp

        return OpenAIMCPServerStreamableHttp(mcp_tool=mcp_tool)
    if agent_framework is AgentFramework.SMOLAGENTS:
        from .smolagents import SmolagentsMCPServerStreamableHttp

        return SmolagentsMCPServerStreamableHttp(mcp_tool=mcp_tool)
    if agent_framework is AgentFramework.TINYAGENT:
        from .tinyagent import TinyAgentMCPServerStreamableHttp

        return TinyAgentMCPServerStreamableHttp(mcp_tool=mcp_tool)
    assert_never(agent_framework)


def _get_mcp_server(mcp_tool: MCPParams, agent_framework: AgentFramework) -> MCPServer:
    if isinstance(mcp_tool, MCPStdio):
        return _get_stdio_mcp_server(mcp_tool, agent_framework)
    if isinstance(mcp_tool, MCPSse):
        return _get_sse_mcp_server(mcp_tool, agent_framework)
    if isinstance(mcp_tool, MCPStreamableHttp):
        return _get_streamablehttp_mcp_server(mcp_tool, agent_framework)
    assert_never(mcp_tool)


__all__ = [
    "MCPServer",
    "_get_mcp_server",
]



from typing import Any

from litellm import acompletion
from litellm.utils import supports_response_schema
from pydantic import BaseModel

from any_agent.config import AgentFramework
from any_agent.evaluation.schemas import EvaluationOutput
from any_agent.utils.asyncio_sync import run_async_in_sync

DEFAULT_PROMPT_TEMPLATE = """Please answer the evaluation question given the following contextual information:

CONTEXT:
{context}

EVALUATION QUESTION:
{question}"""


LLM_JUDGE_SYSTEM_PROMPT = """You are an expert evaluator that analyzes contextual information to answer specific questions about agent performance and behavior.

You will be provided with:
1. Contextual information of an agent's execution that may be relevant to the evaluation question
2. A specific evaluation question to answer

Your task is to carefully analyze the context and provide a judgment on whether the agent's performance meets the criteria specified in the question.

EVALUATION GUIDELINES:
- Be objective and thorough in your analysis
- If the question asks about specific actions, look for evidence of those actions in the context
- If unsure, err on the side of being more critical rather than lenient

Your output must match the following JSON schema:
{response_schema}"""


class LlmJudge:
    def __init__(
        self,
        model_id: str,
        framework: AgentFramework = AgentFramework.TINYAGENT,
        output_type: type[BaseModel] = EvaluationOutput,
        model_args: dict[str, Any] | None = None,
        system_prompt: str = LLM_JUDGE_SYSTEM_PROMPT,
    ):
        if model_args is None:
            model_args = {}
        self.model_id = model_id
        self.framework = framework
        self.model_args = model_args
        self.output_type = output_type
        self.system_prompt = system_prompt.format(
            response_schema=self.output_type.model_json_schema()
        )
        # If LiteLLM detects that the model supports response_format, set it to the output_type automatically
        if supports_response_schema(model=self.model_id):
            self.model_args["response_format"] = self.output_type

    def _create_prompt(self, context: str, question: str, prompt: str) -> str:
        if "{context}" not in prompt or "{question}" not in prompt:
            msg = "Prompt must contain the following placeholders: {context} and {question}"
            raise ValueError(msg)
        return prompt.format(
            context=context,
            question=question,
        )

    def run(
        self,
        context: str,
        question: str,
        prompt_template: str = DEFAULT_PROMPT_TEMPLATE,
    ) -> BaseModel:
        """Run the judge synchronously.

        Args:
            context: Any relevant information that may be needed to answer the question
            question: The question to ask the agent
            prompt_template: The prompt to use for the LLM

        Returns:
            The evaluation result

        """
        return run_async_in_sync(self.run_async(context, question, prompt_template))

    async def run_async(
        self,
        context: str,
        question: str,
        prompt_template: str = DEFAULT_PROMPT_TEMPLATE,
    ) -> BaseModel:
        """Run the LLM asynchronously.

        Args:
            context: Any relevant information that may be needed to answer the question
            question: The question to ask the agent
            prompt_template: The prompt to use for the LLM

        Returns:
            The evaluation result

        """
        prompt = self._create_prompt(context, question, prompt_template)

        # Make the LLM call
        response = await acompletion(
            model=self.model_id,
            messages=[
                {"role": "system", "content": self.system_prompt},
                {"role": "user", "content": prompt},
            ],
            **self.model_args,
        )

        return self.output_type.model_validate_json(
            response.choices[0].message["content"]
        )



from .a2a import a2a_tool, a2a_tool_async
from .final_output import prepare_final_output
from .mcp import (
    MCPServer,
    _get_mcp_server,
    _MCPConnection,
    _MCPServerBase,
)
from .user_interaction import (
    ask_user_verification,
    send_console_message,
    show_final_output,
    show_plan,
)
from .web_browsing import search_tavily, search_web, visit_webpage

__all__ = [
    "MCPServer",
    "_MCPConnection",
    "_MCPServerBase",
    "_get_mcp_server",
    "a2a_tool",
    "a2a_tool_async",
    "ask_user_verification",
    "prepare_final_output",
    "search_tavily",
    "search_web",
    "send_console_message",
    "show_final_output",
    "show_plan",
    "visit_webpage",
]



# mypy: disable-error-code="method-assign,no-untyped-call,no-untyped-def,union-attr"
from __future__ import annotations

from typing import TYPE_CHECKING, Any

from any_agent.callbacks.span_generation.base import _SpanGeneration

if TYPE_CHECKING:
    from litellm.types.utils import ChatCompletionMessageToolCall, Usage
    from llama_index.core.agent.workflow.workflow_events import AgentOutput
    from llama_index.core.base.llms.types import ChatMessage
    from llama_index.core.tools import ToolMetadata

    from any_agent.callbacks.context import Context


class _LlamaIndexSpanGeneration(_SpanGeneration):
    def before_llm_call(self, context: Context, *args, **kwargs) -> Context:
        # Handle direct dict format (from call_model wrapper)
        if "messages" in kwargs and isinstance(kwargs["messages"], list):
            input_messages = kwargs["messages"]
            if not input_messages:
                return context
            model_id = kwargs.get("model", "No model")
        # Handle LlamaIndex callback format
        elif len(args) >= 2 and isinstance(args[1], list):
            llm_input: list[ChatMessage] = args[1]
            input_messages = [
                {
                    "role": message.role.value,
                    "content": message.content or "No content",
                }
                for message in llm_input
            ]
            model_id = context.shared["model_id"]
        else:
            return context

        return self._set_llm_input(context, model_id, input_messages)

    def after_llm_call(self, context: Context, *args, **kwargs) -> Context:
        response = args[0]
        token_usage: Usage | None
        # Handle litellm ModelResponse (from call_model wrapper)
        if hasattr(response, "choices") and hasattr(response, "model_extra"):
            if not response.choices:
                return context

            message = getattr(response.choices[0], "message", None)
            if not message:
                return context

            output: str | list[dict[str, Any]] = ""
            if content := getattr(message, "content", None):
                output = content

            tool_calls: list[ChatCompletionMessageToolCall] | None
            if tool_calls := getattr(message, "tool_calls", None):
                output = [
                    {
                        "tool.name": getattr(tool_call.function, "name", "No name"),
                        "tool.args": getattr(
                            tool_call.function, "arguments", "No name"
                        ),
                    }
                    for tool_call in tool_calls
                    if tool_call.function
                ]

            input_tokens = 0
            output_tokens = 0

            if token_usage := getattr(response, "model_extra", {}).get("usage"):
                if token_usage:
                    input_tokens = token_usage.prompt_tokens
                    output_tokens = token_usage.completion_tokens

        # Handle LlamaIndex AgentOutput
        elif hasattr(response, "response") or hasattr(response, "tool_calls"):
            agent_output: AgentOutput = response

            output = ""
            if agent_response := agent_output.response:
                if content := agent_response.content:
                    output = content

            # Fix type annotation issue - agent_output.tool_calls returns different type than ChatCompletionMessageToolCall
            if agent_tool_calls := agent_output.tool_calls:
                output = [
                    {
                        "tool.name": getattr(tool_call, "tool_name", "No name"),
                        "tool.args": getattr(tool_call, "tool_kwars", "{}"),
                    }
                    for tool_call in agent_tool_calls
                ]

            input_tokens = 0
            output_tokens = 0
            raw: dict[str, Any] | None
            if raw := getattr(agent_output, "raw", None):
                # Rename to avoid variable redefinition
                usage_info: dict[str, int]
                if usage_info := raw.get("usage"):
                    input_tokens = usage_info.get("prompt_tokens", 0)
                    output_tokens = usage_info.get("completion_tokens", 0)
        else:
            return context

        return self._set_llm_output(context, output, input_tokens, output_tokens)

    def before_tool_execution(self, context: Context, *args, **kwargs) -> Context:
        meta: ToolMetadata = context.shared["metadata"]

        return self._set_tool_input(
            context, name=str(meta.name), description=meta.description, args=kwargs
        )

    def after_tool_execution(self, context: Context, *args, **kwargs) -> Context:
        output = args[0]

        if raw_output := getattr(output, "raw_output", None):
            if content := getattr(raw_output, "content", None):
                return self._set_tool_output(context, content[0].text)
            return self._set_tool_output(context, raw_output)
        return self._set_tool_output(context, output)



# mypy: disable-error-code="arg-type,attr-defined"
from __future__ import annotations

import json
from datetime import timedelta
from functools import cached_property
from typing import TYPE_CHECKING, Any, Literal

from opentelemetry.sdk.trace import ReadableSpan
from pydantic import BaseModel, ConfigDict, Field

from any_agent.logging import logger

from .attributes import GenAI
from .otel_types import (
    AttributeValue,
    Event,
    Link,
    Resource,
    SpanContext,
    SpanKind,
    Status,
)

if TYPE_CHECKING:
    from collections.abc import Mapping

    from opentelemetry.trace import Span


class TokenInfo(BaseModel):
    """Token Count information."""

    input_tokens: int
    """Number of input tokens."""

    output_tokens: int
    """Number of output tokens."""

    @property
    def total_tokens(self) -> int:
        """Total number of tokens."""
        return self.input_tokens + self.output_tokens

    model_config = ConfigDict(extra="forbid")


class CostInfo(BaseModel):
    """Cost information."""

    input_cost: float
    "Cost associated to the input tokens."

    output_cost: float
    """Cost associated to the output tokens."""

    @property
    def total_cost(self) -> float:
        """Total cost."""
        return self.input_cost + self.output_cost

    model_config = ConfigDict(extra="forbid")


class AgentMessage(BaseModel):
    """A message that can be exported to JSON or printed to the console."""

    role: Literal["user", "assistant", "system", "tool"]
    content: str

    model_config = ConfigDict(extra="forbid")


class AgentSpan(BaseModel):
    """A span that can be exported to JSON or printed to the console."""

    name: str
    kind: SpanKind
    parent: SpanContext | None = None
    start_time: int | None = None
    end_time: int | None = None
    status: Status
    context: SpanContext
    attributes: dict[str, Any]
    links: list[Link]
    events: list[Event]
    resource: Resource

    model_config = ConfigDict(arbitrary_types_allowed=False)

    @classmethod
    def from_otel(cls, otel_span: Span) -> AgentSpan:
        """Create an AgentSpan from an OTEL Span."""
        return cls(
            name=otel_span.name,
            kind=SpanKind.from_otel(otel_span.kind),
            parent=SpanContext.from_otel(otel_span.parent),
            start_time=otel_span.start_time,
            end_time=otel_span.end_time,
            status=Status.from_otel(otel_span.status),
            context=SpanContext.from_otel(otel_span.context),
            attributes=dict(otel_span.attributes) if otel_span.attributes else {},
            links=[Link.from_otel(link) for link in otel_span.links],
            events=[Event.from_otel(event) for event in otel_span.events],
            resource=Resource.from_otel(otel_span.resource),
        )

    def to_readable_span(self) -> ReadableSpan:
        """Create an ReadableSpan from the AgentSpan."""
        return ReadableSpan(
            name=self.name,
            kind=self.kind,
            parent=self.parent,
            start_time=self.start_time,
            end_time=self.end_time,
            status=self.status,
            context=self.context,
            attributes=self.attributes,
            links=self.links,
            events=self.events,
            resource=self.resource,
        )

    def set_attributes(self, attributes: Mapping[str, AttributeValue]) -> None:
        """Set attributes for the span."""
        for key, value in attributes.items():
            if key in self.attributes:
                logger.warning("Overwriting attribute %s with %s", key, value)
            self.attributes[key] = value

    def is_agent_invocation(self) -> bool:
        """Check whether this span is an agent invocation (the very first span)."""
        return self.attributes.get(GenAI.OPERATION_NAME) == "invoke_agent"

    def is_llm_call(self) -> bool:
        """Check whether this span is a call to an LLM."""
        return self.attributes.get(GenAI.OPERATION_NAME) == "call_llm"

    def is_tool_execution(self) -> bool:
        """Check whether this span is an execution of a tool."""
        return self.attributes.get(GenAI.OPERATION_NAME) == "execute_tool"

    def get_input_messages(self) -> list[AgentMessage] | None:
        """Extract input messages from an LLM call span.

        Returns:
            List of message dicts with 'role' and 'content' keys, or None if not available.

        """
        if not self.is_llm_call():
            msg = "Span is not an LLM call"
            raise ValueError(msg)

        messages_json = self.attributes.get(GenAI.INPUT_MESSAGES)
        if not messages_json:
            logger.debug("No input messages found in span")
            return None

        try:
            parsed_messages = json.loads(messages_json)
            # Ensure it's a list of dicts
        except (json.JSONDecodeError, TypeError) as e:
            msg = "Failed to parse input messages from span"
            logger.error(msg)
            raise ValueError(msg) from e
        if not isinstance(parsed_messages, list):
            msg = "Input messages are not a list of messages"
            raise ValueError(msg)
        return [AgentMessage.model_validate(msg) for msg in parsed_messages]

    def get_output_content(self) -> str | None:
        """Extract output content from an LLM call or tool execution span.

        Returns:
            The output content as a string, or None if not available.

        """
        if not self.is_llm_call() and not self.is_tool_execution():
            msg = "Span is not an LLM call or tool execution"
            raise ValueError(msg)

        output = self.attributes.get(GenAI.OUTPUT)
        if not output:
            logger.debug("No output found in span")
            return None
        return str(output)


class AgentTrace(BaseModel):
    """A trace that can be exported to JSON or printed to the console."""

    spans: list[AgentSpan] = Field(default_factory=list)
    """A list of [`AgentSpan`][any_agent.tracing.agent_trace.AgentSpan] that form the trace.
    """

    final_output: str | dict[str, Any] | BaseModel | None = Field(default=None)
    """Contains the final output message returned by the agent.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def _invalidate_tokens_and_cost_cache(self) -> None:
        """Clear the cached tokens_and_cost property if it exists."""
        if "tokens" in self.__dict__:
            del self.tokens
        if "cost" in self.__dict__:
            del self.cost

    def add_span(self, span: AgentSpan | Span) -> None:
        """Add an AgentSpan to the trace and clear the tokens_and_cost cache if present."""
        if not isinstance(span, AgentSpan):
            span = AgentSpan.from_otel(span)
        self.spans.append(span)
        self._invalidate_tokens_and_cost_cache()

    def add_spans(self, spans: list[AgentSpan]) -> None:
        """Add a list of AgentSpans to the trace and clear the tokens_and_cost cache if present."""
        self.spans.extend(spans)
        self._invalidate_tokens_and_cost_cache()

    def spans_to_messages(self) -> list[AgentMessage]:
        """Convert spans to standard message format.

        Returns:
            List of message dicts with 'role' and 'content' keys.

        """
        messages: list[AgentMessage] = []

        # Process spans in chronological order (excluding the final invoke_agent span)
        # Filter out any agent invocation spans
        filtered_spans: list[AgentSpan] = []
        for span in self.spans:
            if not span.is_agent_invocation():
                filtered_spans.append(span)

        for span in filtered_spans:
            if span.is_llm_call():
                # Extract input messages from the span
                input_messages = span.get_input_messages()
                if input_messages:
                    for msg in input_messages:
                        if not any(
                            existing.role == msg.role
                            and existing.content == msg.content
                            for existing in messages
                        ):
                            messages.append(msg)

                # Add the assistant's response
                output_content = span.get_output_content()
                if output_content:
                    # Avoid duplicate assistant messages
                    if not (
                        messages
                        and messages[-1].role == "assistant"
                        and messages[-1].content == output_content
                    ):
                        messages.append(
                            AgentMessage(role="assistant", content=output_content)
                        )

            elif span.is_tool_execution():
                # For tool executions, include the result in the conversation
                output_content = span.get_output_content()
                if output_content:
                    tool_name = span.attributes[GenAI.TOOL_NAME]
                    tool_args = span.attributes.get(GenAI.TOOL_ARGS, "{}")
                    messages.append(
                        AgentMessage(
                            role="assistant",
                            content=f"[Tool {tool_name} executed: {output_content} with args: {tool_args}]",
                        )
                    )

        return messages

    @property
    def duration(self) -> timedelta:
        """Duration of the parent `invoke_agent` span as a datetime.timedelta object.

        The duration is computed from the span's start and end time (in nanoseconds).

        Raises ValueError if:
            - There are no spans.
            - The invoke_agent span is not the last span.
            - Any of the start/end times are missing.
        """
        if not self.spans:
            msg = "No spans found in trace"
            raise ValueError(msg)
        span = self.spans[-1]
        if not span.is_agent_invocation():
            msg = "Last span is not `invoke_agent`"
            raise ValueError(msg)
        if span.start_time is not None and span.end_time is not None:
            duration_ns = span.end_time - span.start_time
            return timedelta(seconds=duration_ns / 1_000_000_000)
        msg = "Start or end time is missing for the `invoke_agent` span"
        raise ValueError(msg)

    @cached_property
    def tokens(self) -> TokenInfo:
        """The [`TokenInfo`][any_agent.tracing.agent_trace.TokenInfo] for this trace. Cached after first computation."""
        sum_input_tokens = 0
        sum_output_tokens = 0
        for span in self.spans:
            if span.is_llm_call():
                sum_input_tokens += span.attributes.get(GenAI.USAGE_INPUT_TOKENS, 0)
                sum_output_tokens += span.attributes.get(GenAI.USAGE_OUTPUT_TOKENS, 0)
        return TokenInfo(input_tokens=sum_input_tokens, output_tokens=sum_output_tokens)

    @cached_property
    def cost(self) -> CostInfo:
        """The [`CostInfo`][any_agent.tracing.agent_trace.CostInfo] for this trace. Cached after first computation."""
        sum_input_cost = 0.0
        sum_output_cost = 0.0
        for span in self.spans:
            if span.is_llm_call():
                sum_input_cost += span.attributes.get(GenAI.USAGE_INPUT_COST, 0)
                sum_output_cost += span.attributes.get(GenAI.USAGE_OUTPUT_COST, 0)
        return CostInfo(input_cost=sum_input_cost, output_cost=sum_output_cost)



import pytest

# Skip entire module if a2a dependencies are not available
pytest.importorskip("a2a")


@pytest.mark.asyncio
async def test_a2a_imports() -> None:
    """Test that A2A serving can be imported."""
    from any_agent.serving import serve_a2a_async



# mypy: disable-error-code="method-assign,no-untyped-def,union-attr"
from __future__ import annotations

from typing import TYPE_CHECKING, Any

from any_agent.callbacks.span_generation.base import _SpanGeneration

if TYPE_CHECKING:
    from agno.models.message import Message, MessageMetrics
    from agno.tools.function import FunctionCall

    from any_agent.callbacks.context import Context


class _AgnoSpanGeneration(_SpanGeneration):
    def before_llm_call(self, context: Context, *args, **kwargs):
        messages: list[Message] = kwargs.get("messages", [])
        input_messages = [
            {"role": message.role, "content": str(message.content)}
            for message in messages
        ]
        return self._set_llm_input(context, context.shared["model_id"], input_messages)

    def after_llm_call(self, context: Context, *args, **kwargs) -> Context:
        output: str | list[dict[str, Any]] = ""
        if assistant_message := kwargs.get("assistant_message"):
            if content := getattr(assistant_message, "content", None):
                output = str(content)
            if tool_calls := getattr(assistant_message, "tool_calls", None):
                output = [
                    {
                        "tool.name": tool.get("function", {}).get("name", "No name"),
                        "tool.args": tool.get("function", {}).get(
                            "arguments", "No args"
                        ),
                    }
                    for tool in tool_calls
                ]

            metrics: MessageMetrics | None
            input_tokens: int = 0
            output_tokens: int = 0
            if metrics := getattr(assistant_message, "metrics", None):
                input_tokens = metrics.input_tokens
                output_tokens = metrics.output_tokens

            context = self._set_llm_output(context, output, input_tokens, output_tokens)

        return context

    def before_tool_execution(self, context: Context, *args, **kwargs) -> Context:
        function_call: FunctionCall = args[0]
        function = function_call.function

        return self._set_tool_input(
            context,
            name=function.name,
            description=function.description,
            args=function_call.arguments,
            call_id=function_call.call_id,
        )

    def after_tool_execution(self, context: Context, *args, **kwargs) -> Context:
        function_call: FunctionCall = args[1]
        return self._set_tool_output(context, function_call.result)



from collections.abc import Sequence
from typing import Any

import pytest
from agents.tool import Tool as OpenaiClass
from google.adk.tools import FunctionTool as GoogleClass
from langchain_core.tools import BaseTool as LangchainClass
from llama_index.core.tools import FunctionTool as LlamaindexClass
from smolagents.tools import Tool as SmolagentsClass

from any_agent import AgentFramework
from any_agent.config import Tool
from any_agent.tools import search_web, visit_webpage
from any_agent.tools.wrappers import _wrap_tools
from any_agent.utils import run_async_in_sync


def wrap_sync(
    tools: Sequence[Tool],
    framework: AgentFramework,
) -> list[Tool]:
    wrapped_tools, _ = run_async_in_sync(_wrap_tools(tools, framework))
    return wrapped_tools


@pytest.mark.parametrize(
    ("framework", "expected_class"),
    [
        (AgentFramework.GOOGLE, GoogleClass),
        (AgentFramework.LANGCHAIN, LangchainClass),
        (AgentFramework.LLAMA_INDEX, LlamaindexClass),
        (AgentFramework.OPENAI, OpenaiClass),
        (AgentFramework.SMOLAGENTS, SmolagentsClass),
    ],
)
def test_wrap_tools(framework: AgentFramework, expected_class: Any) -> None:
    wrapped_tools = wrap_sync([search_web, visit_webpage], framework)
    assert all(isinstance(tool, expected_class) for tool in wrapped_tools)



# pylint: disable=unused-argument, unused-variable, attr-de
from collections.abc import Generator, Sequence
from unittest.mock import MagicMock, patch

import pytest

from any_agent.config import AgentFramework, MCPSse, MCPStdio, MCPStreamableHttp, Tool
from any_agent.tools import _get_mcp_server


@pytest.fixture
def smolagents_mcp_server(
    tools: Sequence[Tool],
) -> Generator[MagicMock]:
    with patch(
        "any_agent.tools.mcp.frameworks.smolagents.MCPClient"
    ) as mock_client_class:
        mock_client_class.return_value.__enter__.return_value = tools
        yield mock_client_class


@pytest.mark.asyncio
async def test_smolagents_mcp_sse_integration(
    mcp_sse_params_no_tools: MCPSse,
    smolagents_mcp_server: MagicMock,
) -> None:
    server = _get_mcp_server(mcp_sse_params_no_tools, AgentFramework.SMOLAGENTS)

    await server._setup_tools()

    smolagents_mcp_server.assert_called_once_with(
        {"url": mcp_sse_params_no_tools.url, "transport": "sse"},
        adapter_kwargs={
            "client_session_timeout_seconds": mcp_sse_params_no_tools.client_session_timeout_seconds
        },
    )


@pytest.mark.asyncio
async def test_smolagents_mcp_streamablehttp_integration(
    mcp_streamablehttp_params_no_tools: MCPStreamableHttp,
    smolagents_mcp_server: MagicMock,
) -> None:
    server = _get_mcp_server(
        mcp_streamablehttp_params_no_tools, AgentFramework.SMOLAGENTS
    )

    await server._setup_tools()

    smolagents_mcp_server.assert_called_once_with(
        {"url": mcp_streamablehttp_params_no_tools.url, "transport": "streamable-http"},
        adapter_kwargs={
            "client_session_timeout_seconds": mcp_streamablehttp_params_no_tools.client_session_timeout_seconds
        },
    )


@pytest.mark.asyncio
async def test_smolagents_mcp_env() -> None:
    mcp_server = _get_mcp_server(
        MCPStdio(command="print('Hello MCP')", args=[], env={"FOO": "BAR"}),
        AgentFramework.SMOLAGENTS,
    )
    mocked_class = MagicMock()

    with (
        patch("any_agent.tools.mcp.frameworks.smolagents.MCPClient", mocked_class),
    ):
        await mcp_server._setup_tools()
        assert mocked_class.call_args_list[0][0][0].env == {"FOO": "BAR"}


@pytest.mark.asyncio
async def test_smolagents_mcp_stdio_timeout() -> None:
    """Test that the timeout is properly passed to MCPClient for STDIO connections."""
    custom_timeout = 10.0
    mcp_server = _get_mcp_server(
        MCPStdio(
            command="test_command",
            args=[],
            env={},
            client_session_timeout_seconds=custom_timeout,
        ),
        AgentFramework.SMOLAGENTS,
    )
    mocked_class = MagicMock()

    with patch("any_agent.tools.mcp.frameworks.smolagents.MCPClient", mocked_class):
        await mcp_server._setup_tools()
        call_args = mocked_class.call_args_list[0]
        assert (
            call_args[1]["adapter_kwargs"]["client_session_timeout_seconds"]
            == custom_timeout
        )


@pytest.mark.asyncio
async def test_smolagents_mcp_sse_timeout() -> None:
    """Test that the timeout is properly passed to MCPClient for SSE connections."""
    custom_timeout = 15.0
    mcp_sse_params = MCPSse(
        url="http://localhost:8000/sse",
        client_session_timeout_seconds=custom_timeout,
    )

    with patch("any_agent.tools.mcp.frameworks.smolagents.MCPClient") as mock_client:
        mcp_server = _get_mcp_server(mcp_sse_params, AgentFramework.SMOLAGENTS)
        await mcp_server._setup_tools()

        mock_client.assert_called_once()
        call_args = mock_client.call_args_list[0]
        assert call_args[0][0] == {"url": mcp_sse_params.url, "transport": "sse"}
        assert (
            call_args[1]["adapter_kwargs"]["client_session_timeout_seconds"]
            == custom_timeout
        )



# mypy: disable-error-code="no-untyped-def,union-attr"
from __future__ import annotations

from typing import TYPE_CHECKING, Any

from any_agent.callbacks.span_generation.base import _SpanGeneration

if TYPE_CHECKING:
    from smolagents.models import ChatMessage
    from smolagents.tools import Tool

    from any_agent.callbacks.context import Context


class _SmolagentsSpanGeneration(_SpanGeneration):
    def before_llm_call(self, context: Context, *args, **kwargs) -> Context:
        model_id = context.shared["model_id"]

        messages: list[ChatMessage] = args[0]
        input_messages = [
            {
                "role": message.role.value,
                "content": message.content[0]["text"],  # type: ignore[index]
            }
            for message in messages
            if message.content
        ]

        return self._set_llm_input(context, model_id, input_messages)

    def after_llm_call(self, context: Context, *args, **kwargs) -> Context:
        response: ChatMessage = args[0]
        output: str | list[dict[str, Any]]
        if content := response.content:
            output = str(content)
        elif tool_calls := response.tool_calls:
            output = [
                {
                    "tool.name": tool_call.function.name,
                    "tool.args": tool_call.function.arguments,
                }
                for tool_call in tool_calls
            ]

        input_tokens = 0
        output_tokens = 0
        if raw := response.raw:
            if token_usage := raw.get("usage", None):
                input_tokens = token_usage.prompt_tokens
                output_tokens = token_usage.completion_tokens

        return self._set_llm_output(context, output, input_tokens, output_tokens)

    def before_tool_execution(self, context: Context, *args, **kwargs) -> Context:
        tool: Tool = context.shared["original_tool"]

        return self._set_tool_input(
            context, name=tool.name, description=tool.description, args=kwargs
        )

    def after_tool_execution(self, context: Context, *args, **kwargs) -> Context:
        return self._set_tool_output(context, args[0])



import logging
from collections.abc import Generator
from typing import Any
from unittest.mock import patch

import pytest


@pytest.fixture
def local_logger(monkeypatch: Any) -> Generator[logging.Logger, None, None]:
    # Create a fresh logger instance for each test
    test_logger = logging.getLogger(f"test_logger_{id(object())}")
    # Remove all handlers
    for handler in test_logger.handlers[:]:
        test_logger.removeHandler(handler)
    test_logger.handlers.clear()
    test_logger.setLevel(logging.NOTSET)
    test_logger.propagate = False
    # Patch the logger in both import styles
    with patch("any_agent.logging.logger", test_logger):
        yield test_logger



from collections.abc import Generator, Sequence
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from llama_index.tools.mcp import BasicMCPClient as LlamaIndexMCPClient
from llama_index.tools.mcp import McpToolSpec as LlamaIndexMcpToolSpec

from any_agent.config import AgentFramework, MCPSse, MCPStdio, Tool
from any_agent.tools import _get_mcp_server


@pytest.fixture
def llama_index_mcp_client() -> Generator[LlamaIndexMCPClient, None, None]:
    with patch(
        "any_agent.tools.mcp.frameworks.llama_index.LlamaIndexMCPClient"
    ) as mock_client:
        yield mock_client


@pytest.fixture
def llama_index_mcp_tool_spec(
    tools: Sequence[Tool],
) -> Generator[LlamaIndexMcpToolSpec, None, None]:
    tool_spec = MagicMock()
    tool_spec.to_tool_list_async = AsyncMock(return_value=tools)
    with patch(
        "any_agent.tools.mcp.frameworks.llama_index.LlamaIndexMcpToolSpec",
        return_value=tool_spec,
    ) as mock_tool_spec:
        yield mock_tool_spec


@pytest.mark.asyncio
@pytest.mark.usefixtures(
    "llama_index_mcp_tool_spec",
)
async def test_llamaindex_mcp_sse_integration(
    mcp_sse_params_with_tools: MCPSse,
    llama_index_mcp_client: LlamaIndexMCPClient,
) -> None:
    server = _get_mcp_server(mcp_sse_params_with_tools, AgentFramework.LLAMA_INDEX)

    await server._setup_tools()

    llama_index_mcp_client.assert_called_once_with(  # type: ignore[attr-defined]
        command_or_url=mcp_sse_params_with_tools.url,
        timeout=int(mcp_sse_params_with_tools.client_session_timeout_seconds)
        if mcp_sse_params_with_tools.client_session_timeout_seconds is not None
        else 5,
    )


@pytest.mark.asyncio
async def test_llama_index_mcp_env() -> None:
    mcp_server = _get_mcp_server(
        MCPStdio(command="print('Hello MCP')", args=[], env={"FOO": "BAR"}),
        AgentFramework.LLAMA_INDEX,
    )
    mocked_class = MagicMock()
    mocked_class.return_value = AsyncMock()

    with (
        patch(
            "any_agent.tools.mcp.frameworks.llama_index.LlamaIndexMcpToolSpec",
            mocked_class,
        ),
    ):
        await mcp_server._setup_tools()
        assert mocked_class.call_args_list[0][1]["client"].env == {"FOO": "BAR"}


@pytest.mark.asyncio
async def test_llamaindex_client_session_timeout_passed() -> None:
    """Test that client_session_timeout_seconds parameter is properly passed to LlamaIndex BasicMCPClient (STDIO only)."""
    custom_timeout = 15.0
    stdio_params = MCPStdio(
        command="echo",
        args=["test"],
        client_session_timeout_seconds=custom_timeout,
    )
    sse_params = MCPSse(
        url="http://localhost:8000",
        client_session_timeout_seconds=custom_timeout,
    )
    # STDIO
    server = _get_mcp_server(stdio_params, AgentFramework.LLAMA_INDEX)
    with patch(
        "any_agent.tools.mcp.frameworks.llama_index.LlamaIndexMCPClient"
    ) as mock_client:
        mock_client_instance = MagicMock()
        mock_client.return_value = mock_client_instance
        with patch(
            "any_agent.tools.mcp.frameworks.llama_index.LlamaIndexMcpToolSpec"
        ) as mock_spec:
            mock_spec_instance = MagicMock()
            mock_spec_instance.to_tool_list_async = AsyncMock(return_value=[])
            mock_spec.return_value = mock_spec_instance
            await server._setup_tools()
            mock_client.assert_called_once()
            call_args = mock_client.call_args
            assert call_args.kwargs["timeout"] == custom_timeout
    # SSE (check that timeout is passed to LlamaIndexMCPClient)
    server = _get_mcp_server(sse_params, AgentFramework.LLAMA_INDEX)
    with patch(
        "any_agent.tools.mcp.frameworks.llama_index.LlamaIndexMCPClient"
    ) as mock_client:
        mock_client_instance = MagicMock()
        mock_client.return_value = mock_client_instance
        with patch(
            "any_agent.tools.mcp.frameworks.llama_index.LlamaIndexMcpToolSpec"
        ) as mock_spec:
            mock_spec_instance = MagicMock()
            mock_spec_instance.to_tool_list_async = AsyncMock(return_value=[])
            mock_spec.return_value = mock_spec_instance
            await server._setup_tools()
            mock_client.assert_called_once()
            call_args = mock_client.call_args
            assert call_args.kwargs["timeout"] == custom_timeout



# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
from typing import Any
from unittest.mock import patch

import pytest

from any_agent import AgentConfig, AgentFramework, AnyAgent
from any_agent.testing.helpers import LITELLM_IMPORT_PATHS

TEST_TEMPERATURE = 0.54321
TEST_PENALTY = 0.5
TEST_QUERY = "what's the state capital of Pennsylvania"
EXPECTED_OUTPUT = "The state capital of Pennsylvania is Harrisburg."


def create_agent_with_model_args(framework: AgentFramework) -> AnyAgent:
    """Helper function to create an agent with test model arguments"""
    return AnyAgent.create(
        framework,
        AgentConfig(
            model_id="mistral/mistral-small-latest",
            model_args={
                "temperature": TEST_TEMPERATURE,
                "frequency_penalty": TEST_PENALTY,
            },
        ),
    )


def test_create_any_with_framework(agent_framework: AgentFramework) -> None:
    agent = AnyAgent.create(
        agent_framework, AgentConfig(model_id="mistral/mistral-small-latest")
    )
    assert agent


def test_create_any_with_valid_string(agent_framework: AgentFramework) -> None:
    agent = AnyAgent.create(
        agent_framework.name, AgentConfig(model_id="mistral/mistral-small-latest")
    )
    assert agent


def test_create_any_with_invalid_string() -> None:
    with pytest.raises(ValueError, match="Unsupported agent framework"):
        AnyAgent.create(
            "non-existing", AgentConfig(model_id="mistral/mistral-small-latest")
        )


def test_model_args(
    agent_framework: AgentFramework,
    mock_litellm_response: Any,
) -> None:
    if agent_framework == AgentFramework.LLAMA_INDEX:
        pytest.skip("LlamaIndex agent uses a litellm streaming syntax")

    agent = create_agent_with_model_args(agent_framework)

    # Patch the appropriate litellm import path for this framework
    import_path = LITELLM_IMPORT_PATHS[agent_framework]
    with patch(import_path, return_value=mock_litellm_response) as mock_litellm:
        # Run the agent
        result = agent.run(TEST_QUERY)

        # Verify results
        assert EXPECTED_OUTPUT == result.final_output
        assert mock_litellm.call_args.kwargs["temperature"] == TEST_TEMPERATURE
        assert mock_litellm.call_args.kwargs["frequency_penalty"] == TEST_PENALTY
        assert mock_litellm.call_count > 0


def test_model_args_streaming(
    agent_framework: AgentFramework, mock_litellm_streaming: Any
) -> None:
    if agent_framework != AgentFramework.LLAMA_INDEX:
        pytest.skip("This test is only for LlamaIndex framework")

    agent = create_agent_with_model_args(agent_framework)

    # Patch the appropriate litellm import path for LlamaIndex
    import_path = LITELLM_IMPORT_PATHS[agent_framework]
    with patch(import_path, side_effect=mock_litellm_streaming) as mock_litellm:
        # Run the agent
        result = agent.run(TEST_QUERY)

        # Verify results
        assert result.final_output
        assert "Harrisburg" in result.final_output
        assert mock_litellm.call_args.kwargs["stream"] is True
        assert mock_litellm.call_args.kwargs["temperature"] == TEST_TEMPERATURE
        assert mock_litellm.call_args.kwargs["frequency_penalty"] == TEST_PENALTY
        assert mock_litellm.call_count > 0



from pydantic import BaseModel, ConfigDict


class MCPServingConfig(BaseModel):
    """Configuration for serving an agent using the Model Context Protocol (MCP).

    Example:
        config = MCPServingConfig(
            port=8080,
            endpoint="/my-agent",
        )

    """

    model_config = ConfigDict(extra="forbid")

    host: str = "localhost"
    """Will be passed as argument to `uvicorn.run`."""

    port: int = 5000
    """Will be passed as argument to `uvicorn.run`."""

    endpoint: str = "/"
    """Will be pass as argument to `Starlette().add_route`"""

    log_level: str = "warning"
    """Will be passed as argument to the `uvicorn` server."""

    version: str = "0.1.0"



# mypy: disable-error-code="misc,method-assign,no-untyped-def,no-untyped-call,union-attr"
from __future__ import annotations

from typing import TYPE_CHECKING, Any

from opentelemetry.trace import get_current_span

if TYPE_CHECKING:
    from collections.abc import Callable

    from any_agent.callbacks.context import Context
    from any_agent.frameworks.openai import OpenAIAgent


class _OpenAIAgentsWrapper:
    def __init__(self) -> None:
        self.callback_context: dict[int, Context] = {}
        self._original_llm_call: Callable[..., Any] | None = None
        self._original_get_all_tools: Callable[..., Any] | None = None
        self._original_invokes: dict[str, Any] = {}

    async def wrap(self, agent: OpenAIAgent) -> None:
        self._original_llm_call = agent._agent.model.get_response

        async def wrapped_llm_call(*args, **kwargs):
            context = self.callback_context[
                get_current_span().get_span_context().trace_id
            ]
            context.shared["model_id"] = getattr(agent._agent.model, "model", None)

            for callback in agent.config.callbacks:
                context = callback.before_llm_call(context, *args, **kwargs)

            output = await self._original_llm_call(*args, **kwargs)

            for callback in agent.config.callbacks:
                context = callback.after_llm_call(
                    context,
                    output,
                )

            return output

        agent._agent.model.get_response = wrapped_llm_call

        # OpenAI dynamically converts the MCP tools into FunctionTools during run.
        self._original_get_all_tools = agent._agent.get_all_tools

        async def wrapped_tool_execution(
            original_tool, original_invoke, *args, **kwargs
        ):
            context = self.callback_context[
                get_current_span().get_span_context().trace_id
            ]
            context.shared["original_tool"] = original_tool

            for callback in agent.config.callbacks:
                context = callback.before_tool_execution(context, *args, **kwargs)

            output = await original_invoke(*args, **kwargs)

            for callback in agent.config.callbacks:
                context = callback.after_tool_execution(
                    context,
                    output,
                )

            return output

        async def wrapped_get_all_tools(run_context):
            all_tools = await self._original_get_all_tools(run_context)

            class WrappedTool:
                def __init__(self, original_tool, original_invoke):
                    self.original_tool = original_tool
                    self.original_invoke = original_invoke

                async def on_invoke_tool(self, *args, **kwargs):
                    return await wrapped_tool_execution(
                        self.original_tool, self.original_invoke, *args, **kwargs
                    )

            for original_tool in all_tools:
                original_invoke = original_tool.on_invoke_tool
                self._original_invokes[original_tool.name] = original_invoke

                wrapped_tool = WrappedTool(original_tool, original_invoke)
                original_tool.on_invoke_tool = wrapped_tool.on_invoke_tool

            return all_tools

        agent._agent.get_all_tools = wrapped_get_all_tools

    async def unwrap(self, agent: OpenAIAgent) -> None:
        if self._original_llm_call is not None:
            agent._agent.model.get_response = self._original_llm_call

        if self._original_get_all_tools is not None:
            agent._agent.get_all_tools = self._original_get_all_tools

        # Only non-mcp tools need to be unwrapped.
        for tool in agent._agent.tools:
            if tool.name in self._original_invokes:
                tool.on_invoke_tool = self._original_invokes[tool.name]



import asyncio
import datetime
from asyncio import Task
from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from types import TracebackType
from typing import TYPE_CHECKING, Any
from uuid import uuid4

import httpx
import pytest
from a2a.client import A2AClient
from a2a.types import MessageSendParams, SendMessageRequest

from any_agent import AnyAgent
from any_agent.serving import A2AServingConfig
from any_agent.testing.helpers import wait_for_server_async

if TYPE_CHECKING:
    from uvicorn.server import Server

# Constants
DEFAULT_TIMEOUT = 10.0
DEFAULT_LONG_TIMEOUT = 1500


class A2ATestHelpers:
    """Helper methods for A2A testing."""

    @staticmethod
    def create_message_payload(
        text: str,
        message_id: str | None = None,
        context_id: str | None = None,
        task_id: str | None = None,
    ) -> dict[str, Any]:
        """Create a standard A2A message payload."""
        payload = {
            "message": {
                "role": "user",
                "parts": [{"kind": "text", "text": text}],
                "messageId": message_id or str(uuid4()),
            }
        }

        if context_id:
            payload["message"]["contextId"] = context_id
        if task_id:
            payload["message"]["taskId"] = task_id

        return payload

    @staticmethod
    def create_send_message_request(
        text: str,
        message_id: str | None = None,
        context_id: str | None = None,
        task_id: str | None = None,
    ) -> SendMessageRequest:
        """Create a SendMessageRequest with standard payload."""
        payload = A2ATestHelpers.create_message_payload(
            text=text,
            message_id=message_id,
            context_id=context_id,
            task_id=task_id,
        )

        return SendMessageRequest(id=str(uuid4()), params=MessageSendParams(**payload))


class A2AServedAgent:
    """Context manager for serving an agent and cleaning it up."""

    def __init__(
        self,
        agent: AnyAgent,
        serving_config: A2AServingConfig | None = None,
    ):
        self.agent = agent
        self.serving_config = serving_config or A2AServingConfig(port=0)
        self.task: Task[Any] | None = None
        self.server: Server | None = None
        self.server_url = ""

    async def __aenter__(self) -> "A2AServedAgent":
        """Start serving the agent."""
        server_handle = await self.agent.serve_async(serving_config=self.serving_config)

        self.task = server_handle.task
        self.server = server_handle.server

        # Get the actual port from the server
        assert self.server is not None
        endpoint = getattr(self.serving_config, "endpoint", "")
        self.server_url = f"http://localhost:{server_handle.port}{endpoint}"

        # Wait for server to be ready
        await wait_for_server_async(self.server_url)

        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Clean up the server."""
        if self.server:
            await self.server.shutdown()
        if self.task:
            self.task.cancel()
            try:
                await self.task
            except asyncio.CancelledError:
                pass


@asynccontextmanager
async def a2a_client_from_agent(
    agent: AnyAgent,
    serving_config: A2AServingConfig | None = None,
    http_timeout: float = DEFAULT_TIMEOUT,
) -> AsyncGenerator[tuple[A2AClient, str], None]:
    """Context manager that serves an agent and provides an A2A client for it."""
    async with A2AServedAgent(agent, serving_config) as served_agent:
        async with httpx.AsyncClient(timeout=http_timeout) as httpx_client:
            client = await A2AClient.get_client_from_agent_card_url(
                httpx_client, served_agent.server_url
            )
            yield client, served_agent.server_url


def get_datetime() -> str:
    """Return the current date and time - common test tool."""
    return str(datetime.datetime.now())


# Common date-related test data
DATE_PROMPT = (
    "What date and time is it right now? "
    "In your answer please include the year, month, day, and time. "
    "Example answer could be something like 'Today is December 15, 2024'"
)


def assert_contains_current_date_info(final_output: str) -> None:
    """Assert that the final output contains current date and time information."""
    now = datetime.datetime.now()
    assert all(
        [
            str(now.year) in final_output,
            str(now.day) in final_output,
            now.strftime("%B") in final_output,
        ]
    )


@pytest.fixture
def a2a_test_helpers() -> type[A2ATestHelpers]:
    """Fixture providing A2A test helper methods."""
    return A2ATestHelpers



from uuid import uuid4

import pytest

# Import your agent and config
from any_agent import AgentConfig, AgentFramework, AnyAgent
from any_agent.serving import A2AServingConfig
from any_agent.testing.helpers import (
    DEFAULT_HTTP_KWARGS,
    DEFAULT_SMALL_MODEL_ID,
    get_default_agent_model_args,
    wait_for_server_async,
)

from .conftest import A2ATestHelpers, a2a_client_from_agent


@pytest.mark.asyncio
async def test_serve_async(test_port: int, a2a_test_helpers: A2ATestHelpers) -> None:
    # Create and serve the agent
    agent = await AnyAgent.create_async(
        "tinyagent",
        AgentConfig(
            model_id=DEFAULT_SMALL_MODEL_ID,
            instructions="Directly answer the question without asking the user for input.",
            description="I'm an agent to help.",
            model_args=get_default_agent_model_args(AgentFramework.TINYAGENT),
        ),
    )

    # Use the context manager for proper cleanup
    async with a2a_client_from_agent(agent, A2AServingConfig(port=test_port)) as (
        client,
        server_url,
    ):
        await wait_for_server_async(server_url)
        request = a2a_test_helpers.create_send_message_request(
            text="What is an agent?",
            message_id=uuid4().hex,
        )
        response = await client.send_message(request, http_kwargs=DEFAULT_HTTP_KWARGS)
        assert response is not None



import asyncio

import pytest
from litellm.utils import validate_environment

from any_agent import AgentConfig, AgentFramework, AnyAgent
from any_agent.testing.helpers import (
    DEFAULT_SMALL_MODEL_ID,
    get_default_agent_model_args,
)


def mock_capital(query: str) -> str:
    """Perform a duckduckgo web search based on your query (think a Google search) then returns the top search results.

    Args:
        query (str): The search query to perform.

    Returns:
        The top search results.

    """
    if "France" in query:
        return "The capital of France is Paris."
    if "Spain" in query:
        return "The capital of Spain is Madrid."
    return "No info"


@pytest.mark.asyncio
async def test_run_agent_concurrently(agent_framework: AgentFramework) -> None:
    """When an agent is run concurrently, state from the first run shouldn't bleed into the second run"""
    model_id = DEFAULT_SMALL_MODEL_ID
    env_check = validate_environment(model_id)
    if not env_check["keys_in_environment"]:
        pytest.skip(f"{env_check['missing_keys']} needed for {agent_framework}")

    agent = await AnyAgent.create_async(
        agent_framework,
        AgentConfig(
            model_id=model_id,
            instructions="You must use the tools to find an answer",
            model_args=get_default_agent_model_args(AgentFramework.TINYAGENT),
            tools=[mock_capital],
        ),
    )
    results = await asyncio.gather(
        agent.run_async("What is the capital of France?"),
        agent.run_async("What is the capital of Spain?"),
    )
    outputs = [r.final_output for r in results]
    assert all(o is not None for o in outputs)

    assert sum("Paris" in str(o) for o in outputs) == 1
    assert sum("Madrid" in str(o) for o in outputs) == 1

    first_spans = results[0].spans
    second_spans = results[1].spans
    assert second_spans[: len(first_spans)] != first_spans, (
        "Spans from the first run should not be in the second"
    )



# mypy: disable-error-code="attr-defined,index,no-untyped-def"
from __future__ import annotations

from typing import TYPE_CHECKING

from litellm.cost_calculator import cost_per_token

from any_agent.callbacks.base import Callback
from any_agent.logging import logger
from any_agent.tracing.attributes import GenAI

if TYPE_CHECKING:
    from collections.abc import Mapping

    from opentelemetry.trace import Span

    from any_agent.callbacks.context import Context
    from any_agent.tracing.otel_types import AttributeValue


def add_cost_info(span: Span) -> None:
    """Use litellm to compute cost and add it to span attributes."""
    attributes: Mapping[str, AttributeValue] = span.attributes
    if any(
        key in attributes
        for key in ["gen_ai.usage.input_tokens", "gen_ai.usage.output_tokens"]
    ):
        try:
            cost_prompt, cost_completion = cost_per_token(
                model=str(attributes.get(GenAI.REQUEST_MODEL, "")),
                prompt_tokens=int(attributes.get(GenAI.USAGE_INPUT_TOKENS, 0)),  # type: ignore[arg-type]
                completion_tokens=int(attributes.get(GenAI.USAGE_OUTPUT_TOKENS, 0)),  # type: ignore[arg-type]
            )
            span.set_attributes(
                {
                    GenAI.USAGE_INPUT_COST: cost_prompt,
                    GenAI.USAGE_OUTPUT_COST: cost_completion,
                }
            )
        except Exception as e:
            msg = f"Error computing cost_per_token: {e}"
            logger.warning(msg)


class AddCostInfo(Callback):
    """Add cost information to the LLM Call spans.

    Extend the LLM Call span attributes with 2 new keys:
        - gen_ai.usage.input_cost
        - gen_ai.usage.output_cost
    """

    def after_llm_call(self, context: Context, *args, **kwargs):
        span = context.current_span
        add_cost_info(span)
        return context



from collections.abc import Callable

from a2a.server.tasks import (
    BasePushNotificationSender,
    InMemoryPushNotificationConfigStore,
    InMemoryTaskStore,
)
from a2a.server.tasks.push_notification_config_store import (
    PushNotificationConfigStore,
)
from a2a.server.tasks.push_notification_sender import PushNotificationSender
from a2a.server.tasks.task_store import TaskStore
from a2a.types import AgentSkill
from pydantic import BaseModel, ConfigDict

from any_agent.tracing.agent_trace import AgentMessage

# Type alias for history formatting function
HistoryFormatter = Callable[[list[AgentMessage], str], str]


def default_history_formatter(messages: list[AgentMessage], current_query: str) -> str:
    """Format conversation history and current query into a single prompt.

    Args:
        messages: List of AgentMessage objects from spans_to_messages()
        current_query: The current user query

    Returns:
        Formatted prompt with conversation history

    """
    if not messages:
        return current_query

    # Convert previous conversation to readable format
    history_text = "\n".join(
        [
            f"{msg.role.capitalize()}: {msg.content}"
            for msg in messages
            if msg.role != "system"
        ]
    )

    return (
        f"Previous conversation:\n{history_text}\n"
        "Please respond to the current user message, taking into account the conversation history above."
        f"Current user message: {current_query}\n"
    )


class A2AServingConfig(BaseModel):
    """Configuration for serving an agent using the Agent2Agent Protocol (A2A).

    Example:
        config = A2AServingConfig(
            port=8080,
            endpoint="/my-agent",
            skills=[
                AgentSkill(
                    id="search",
                    name="web_search",
                    description="Search the web for information"
                )
            ],
            context_timeout_minutes=15
        )

    """

    model_config = ConfigDict(extra="forbid")

    host: str = "localhost"
    """Will be passed as argument to `uvicorn.run`."""

    port: int = 5000
    """Will be passed as argument to `uvicorn.run`."""

    endpoint: str = "/"
    """Will be pass as argument to `Starlette().add_route`"""

    log_level: str = "warning"
    """Will be passed as argument to the `uvicorn` server."""

    skills: list[AgentSkill] | None = None
    """List of skills to be used by the agent.

    If not provided, the skills will be inferred from the tools.
    """

    version: str = "0.1.0"

    context_timeout_minutes: int = 10
    """Context timeout in minutes. Contexts will be cleaned up after this period of inactivity."""

    history_formatter: HistoryFormatter = default_history_formatter
    """Function to format conversation history and current query into a single prompt.
    Takes (messages, current_query) and returns formatted string."""

    task_cleanup_interval_minutes: int = 5
    """Interval in minutes between task cleanup runs."""

    push_notifier_store_type: type[PushNotificationConfigStore] = (
        InMemoryPushNotificationConfigStore
    )

    """Push notifier config store to be used by the agent.

    If not provided, a default in-memory push notifier config store will be used.
    """

    push_notifier_sender_type: type[PushNotificationSender] = BasePushNotificationSender
    """Push notifier sender to be used by the agent.

    If not provided, a default async httpx-based push notifier sender will be used.
    """

    task_store_type: type[TaskStore] = InMemoryTaskStore
    """Task store to be used by the agent.

    If not provided, a default in-memory task store will be used.
    """



"""Generate llms.txt and llms-full.txt files for the documentation.

It also serves markdown files as text files for LLM consumption.
It was initially written in collaboration with Claude 4 Sonnet.
"""

import re
from pathlib import Path

MARKDOWN_EXTENSION = ".md"
BASE_URL = (
    "https://raw.githubusercontent.com/mozilla-ai/any-agent/refs/heads/main/docs/"
)
EXCLUDED_DIRS = {".", "__pycache__"}
MARKDOWN_LINK_PATTERN = r"\[([^\]]+)\]\(([^)]+\.md)\)"
MARKDOWN_LINK_REPLACEMENT = r"[\1](#\2)"


def get_nav_files(nav_config):
    """Extract file paths from mkdocs navigation config in order."""
    files = []

    def extract_files(nav_item):
        if isinstance(nav_item, dict):
            for _, value in nav_item.items():
                if isinstance(value, str):
                    # This is a file reference
                    if value.endswith(MARKDOWN_EXTENSION):
                        files.append(value)
                elif isinstance(value, list):
                    # This is a nested section
                    for item in value:
                        extract_files(item)
        elif isinstance(nav_item, list):
            for item in nav_item:
                extract_files(item)

    extract_files(nav_config)
    return files


def get_ordered_files(nav_config, docs_dir: Path):
    """Get ordered list of markdown files based on navigation and additional files."""
    nav_files = get_nav_files(nav_config)
    all_md_files = [str(f.relative_to(docs_dir)) for f in docs_dir.rglob("*.md")]
    ordered_files = []
    for file in nav_files:
        if file in all_md_files:
            ordered_files.append(file)

    for file in all_md_files:
        if file not in ordered_files:
            ordered_files.append(file)

    return ordered_files


def clean_markdown_content(content, file_path):
    """Clean markdown content for better concatenation."""
    # Remove or replace relative links that won't work in concatenated format
    # Convert relative md links to section references where possible
    content = re.sub(MARKDOWN_LINK_PATTERN, MARKDOWN_LINK_REPLACEMENT, content)

    return f"<!-- Source: {file_path} -->\n\n{content}"


def get_file_description(file_path, docs_dir):
    """Get descriptive text extracted from the actual file."""
    full_path = docs_dir / file_path

    if not full_path.exists():
        return ""

    content = full_path.read_text()
    if content is None:
        return ""

    # Assume subheadings are a good enough description
    # or at least better than the first characters in the file.
    return " - ".join(
        line.replace("## ", "").strip()
        for line in content.split("\n")
        if line.startswith("## ")
    )


def create_file_title(file_path):
    """Create a clean title from file path."""
    if file_path == "index.md":
        return "Introduction"

    return (
        file_path.replace(MARKDOWN_EXTENSION, "")
        .replace("_", " ")
        .replace("/", " - ")
        .title()
    )


def generate_llms_txt(docs_dir, site_dir, nav_config):
    """Generate llms.txt file following llmstxt.org standards."""
    ordered_files = get_ordered_files(nav_config, docs_dir)

    llms_txt_content = []

    llms_txt_content.append("# any-agent")
    llms_txt_content.append("")
    llms_txt_content.append("## Docs")
    llms_txt_content.append("")

    for file_path in ordered_files:
        txt_url = f"{BASE_URL}{file_path}"

        title = create_file_title(file_path)
        description = get_file_description(file_path, docs_dir)

        if description:
            llms_txt_content.append(f"- [{title}]({txt_url}): {description}")
        else:
            llms_txt_content.append(f"- [{title}]({txt_url})")

    llms_txt_dest = site_dir / "llms.txt"
    llms_txt_dest.write_text("\n".join(llms_txt_content))


def generate_llms_full_txt(docs_dir, site_dir, nav_config):
    """Generate llms-full.txt by concatenating all markdown documentation."""
    ordered_files = get_ordered_files(nav_config, docs_dir)

    llms_full_content = []

    llms_full_content.extend(
        [
            "# any-agent Documentation",
            "",
            "> Complete documentation for any-agent - A Python library providing a single interface to different agent frameworks.",
            "",
            "This file contains all documentation pages concatenated for easy consumption by AI systems.",
            "",
            "---",
            "",
        ]
    )

    for file_path in ordered_files:
        full_path = docs_dir / file_path

        if full_path.exists():
            content = full_path.read_text()
            if content is not None:
                cleaned_content = clean_markdown_content(content, file_path)

                # Add section separator
                llms_full_content.extend(
                    [f"## {file_path}", "", cleaned_content, "", "---", ""]
                )

    llms_full_txt_dest = site_dir / "llms-full.txt"
    llms_full_txt_dest.write_text("\n".join(llms_full_content))


def on_post_build(config, **kwargs):
    """Generate llms.txt and llms-full.txt files, and serve markdown as text."""
    docs_dir = Path(config["docs_dir"])
    site_dir = Path(config["site_dir"])

    nav_config = config.get("nav", [])

    generate_llms_txt(docs_dir, site_dir, nav_config)

    generate_llms_full_txt(docs_dir, site_dir, nav_config)



import json
import time
from datetime import datetime, timedelta
from typing import Any

import pytest
from litellm.utils import validate_environment
from pydantic import BaseModel, ConfigDict

from any_agent import (
    AgentConfig,
    AgentFramework,
    AnyAgent,
)
from any_agent.config import MCPStreamableHttp
from any_agent.testing.helpers import (
    DEFAULT_SMALL_MODEL_ID,
    get_default_agent_model_args,
    group_spans,
)
from any_agent.tracing.agent_trace import AgentSpan, AgentTrace, CostInfo, TokenInfo


def assert_trace(agent_trace: AgentTrace, agent_framework: AgentFramework) -> None:
    def assert_first_llm_call(llm_call: AgentSpan) -> None:
        """Checks the `_set_llm_inputs` implemented by each framework's instrumentation."""
        assert llm_call.attributes.get("gen_ai.input.messages", None) is not None
        # input.messages should be a valid JSON string (list of dicts)
        input_messages_raw = llm_call.attributes.get("gen_ai.input.messages")
        assert input_messages_raw is not None
        input_messages = json.loads(input_messages_raw)
        assert input_messages[0]["role"] == "system"
        assert input_messages[1]["role"] == "user"

    def assert_first_tool_execution(tool_execution: AgentSpan) -> None:
        """Checks the tools setup implemented by each framework's instrumentation."""
        assert tool_execution.attributes.get("gen_ai.tool.args", None) is not None
        # tool.args should be a JSON string (dict)
        tool_args_raw = tool_execution.attributes.get("gen_ai.tool.args")
        assert tool_args_raw is not None
        args = json.loads(tool_args_raw)
        assert "timezone" in args
        assert isinstance(agent_trace, AgentTrace)
        assert agent_trace.final_output

    agent_invocations, llm_calls, tool_executions = group_spans(agent_trace.spans)

    assert len(agent_invocations) == 1

    assert len(llm_calls) >= 2
    assert_first_llm_call(llm_calls[0])

    assert len(tool_executions) >= 1
    assert_first_tool_execution(tool_executions[0])

    messages = agent_trace.spans_to_messages()
    assert messages[0].role == "system"
    assert messages[1].role == "user"
    assert len(messages) == 2 + len(llm_calls) + len(tool_executions)


def assert_duration(agent_trace: AgentTrace, wall_time_s: float) -> None:
    assert agent_trace.duration is not None
    assert isinstance(agent_trace.duration, timedelta)
    assert agent_trace.duration.total_seconds() > 0

    diff = abs(agent_trace.duration.total_seconds() - wall_time_s)
    assert diff < 0.1, (
        f"duration ({agent_trace.duration.total_seconds()}s) and wall_time ({wall_time_s}s) differ by more than 0.1s: {diff}s"
    )


def assert_cost(agent_trace: AgentTrace) -> None:
    assert isinstance(agent_trace.cost, CostInfo)
    assert agent_trace.cost.input_cost > 0
    assert agent_trace.cost.output_cost > 0
    assert agent_trace.cost.input_cost + agent_trace.cost.output_cost < 1.00


def assert_tokens(agent_trace: AgentTrace) -> None:
    assert isinstance(agent_trace.tokens, TokenInfo)
    assert agent_trace.tokens.input_tokens > 0
    assert agent_trace.tokens.output_tokens > 0
    assert (agent_trace.tokens.input_tokens + agent_trace.tokens.output_tokens) < 20000


class Step(BaseModel):
    number: int
    description: str


class Steps(BaseModel):
    model_config = ConfigDict(extra="forbid")
    steps: list[Step]


def test_load_and_run_agent_streamable_http(
    agent_framework: AgentFramework,
    request: pytest.FixtureRequest,
    date_streamable_http_server: dict[str, Any],
) -> None:
    kwargs = {}

    kwargs["model_id"] = DEFAULT_SMALL_MODEL_ID
    env_check = validate_environment(kwargs["model_id"])
    if not env_check["keys_in_environment"]:
        pytest.skip(f"{env_check['missing_keys']} needed for {agent_framework}")

    tools = [
        MCPStreamableHttp(
            url=date_streamable_http_server["url"],
            client_session_timeout_seconds=30,
        ),
    ]
    agent_config = AgentConfig(
        tools=tools,
        instructions="Use the available tools to answer.",
        model_args=get_default_agent_model_args(agent_framework),
        output_type=Steps,
        **kwargs,  # type: ignore[arg-type]
    )
    agent = AnyAgent.create(agent_framework, agent_config)

    start_ns = time.time_ns()
    agent_trace = agent.run(
        "Return what year it is in the America/New_York timezone. One of the steps returned must include this year."
    )
    end_ns = time.time_ns()

    assert isinstance(agent_trace.final_output, Steps)

    steps = agent_trace.final_output.steps
    assert any(
        str(datetime.now().year) in steps[n].description for n in range(len(steps))
    )

    assert_trace(agent_trace, agent_framework)
    assert_duration(agent_trace, (end_ns - start_ns) / 1_000_000_000)
    assert_cost(agent_trace)
    assert_tokens(agent_trace)



from syrupy.assertion import SnapshotAssertion

from any_agent.tracing.agent_trace import AgentTrace


def test_agent_trace_snapshot(
    agent_trace: AgentTrace, snapshot: SnapshotAssertion
) -> None:
    # Snapshot the dict representation (so you see changes in the schema)
    # If this assert fails and you decide that you're ok with the new schema,
    # you can easily update the snapshot by running:
    # pytest tests/snapshots --snapshot-update
    assert agent_trace.model_dump() == snapshot(
        name=agent_trace.spans[0].context.trace_id
    )



"""Vendoring https://github.com/Akshay-Dongare/langchain-litellm/blob/main/langchain_litellm/chat_models/litellm.py ."""
# mypy: ignore-errors

from __future__ import annotations

import json
import logging
from collections.abc import AsyncIterator, Callable, Iterator, Mapping, Sequence
from typing import (
    Any,
    Literal,
)

from langchain_core.callbacks import (
    AsyncCallbackManagerForLLMRun,
    CallbackManagerForLLMRun,
)
from langchain_core.language_models import LanguageModelInput
from langchain_core.language_models.chat_models import (
    BaseChatModel,
    agenerate_from_stream,
    generate_from_stream,
)
from langchain_core.language_models.llms import create_base_retry_decorator
from langchain_core.messages import (
    AIMessage,
    AIMessageChunk,
    BaseMessage,
    BaseMessageChunk,
    ChatMessage,
    ChatMessageChunk,
    FunctionMessage,
    FunctionMessageChunk,
    HumanMessage,
    HumanMessageChunk,
    SystemMessage,
    SystemMessageChunk,
    ToolCall,
    ToolCallChunk,
    ToolMessage,
)
from langchain_core.messages.ai import UsageMetadata
from langchain_core.outputs import (
    ChatGeneration,
    ChatGenerationChunk,
    ChatResult,
)
from langchain_core.runnables import Runnable
from langchain_core.tools import BaseTool
from langchain_core.utils import get_from_dict_or_env, pre_init
from langchain_core.utils.function_calling import convert_to_openai_tool
import litellm
from litellm.types.utils import Delta
from pydantic import BaseModel, Field

logger = logging.getLogger(__name__)


class ChatLiteLLMException(Exception):
    """Error with the `LiteLLM I/O` library"""


def _create_retry_decorator(
    llm: ChatLiteLLM,
    run_manager: AsyncCallbackManagerForLLMRun | CallbackManagerForLLMRun | None = None,
) -> Callable[[Any], Any]:
    """Returns a tenacity retry decorator, preconfigured to handle PaLM exceptions"""
    import litellm

    errors = [
        litellm.Timeout,
        litellm.APIError,
        litellm.APIConnectionError,
        litellm.RateLimitError,
    ]
    return create_base_retry_decorator(
        error_types=errors, max_retries=llm.max_retries, run_manager=run_manager
    )


def _convert_dict_to_message(_dict: Mapping[str, Any]) -> BaseMessage:
    role = _dict["role"]
    if role == "user":
        return HumanMessage(content=_dict["content"])
    if role == "assistant":
        # Fix for azure
        # Also OpenAI returns None for tool invocations
        content = _dict.get("content", "") or ""

        additional_kwargs = {}
        if _dict.get("function_call"):
            additional_kwargs["function_call"] = dict(_dict["function_call"])

        if _dict.get("tool_calls"):
            additional_kwargs["tool_calls"] = _dict["tool_calls"]

        return AIMessage(content=content, additional_kwargs=additional_kwargs)
    if role == "system":
        return SystemMessage(content=_dict["content"])
    if role == "function":
        return FunctionMessage(content=_dict["content"], name=_dict["name"])
    if role == "tool":
        return ToolMessage(content=_dict["content"], tool_call_id=_dict["tool_call_id"])
    return ChatMessage(content=_dict["content"], role=role)


async def acompletion_with_retry(
    llm: ChatLiteLLM,
    run_manager: AsyncCallbackManagerForLLMRun | None = None,
    **kwargs: Any,
) -> None:
    """Simple retry wrapper for litellm.acompletion"""
    for try_num in range(llm.max_retries):
        try:
            return await litellm.acompletion(**kwargs)
        except Exception as e:
            logger.error(f"Error in completion with retry: {e}")
            if try_num == llm.max_retries - 1:
                raise e


def _convert_delta_to_message_chunk(
    delta: Delta, default_class: type[BaseMessageChunk]
) -> BaseMessageChunk:
    role = delta.role
    content = delta.content or ""
    if delta.function_call:
        additional_kwargs = {"function_call": dict(delta.function_call)}
    # The hasattr check is necessary because litellm explicitly deletes the
    # `reasoning_content` attribute when it is absent to comply with the OpenAI API.
    # This ensures that the code gracefully handles cases where the attribute is
    # missing, avoiding potential errors or non-compliance with the API.
    elif hasattr(delta, "reasoning_content") and delta.reasoning_content:
        additional_kwargs = {"reasoning_content": delta.reasoning_content}
    else:
        additional_kwargs = {}

    tool_call_chunks = []
    if raw_tool_calls := delta.tool_calls:
        additional_kwargs["tool_calls"] = raw_tool_calls
        try:
            tool_call_chunks = [
                ToolCallChunk(
                    name=rtc.function.name,
                    args=rtc.function.arguments,
                    id=rtc.id,
                    index=rtc.index,
                )
                for rtc in raw_tool_calls
            ]
        except KeyError:
            pass

    if role == "user" or default_class == HumanMessageChunk:
        return HumanMessageChunk(content=content)
    if role == "assistant" or default_class == AIMessageChunk:
        return AIMessageChunk(
            content=content,
            additional_kwargs=additional_kwargs,
            tool_call_chunks=tool_call_chunks,
        )
    if role == "system" or default_class == SystemMessageChunk:
        return SystemMessageChunk(content=content)
    if role == "function" or default_class == FunctionMessageChunk:
        return FunctionMessageChunk(
            content=delta.function_call.arguments, name=delta.function_call.name
        )
    if role or default_class == ChatMessageChunk:
        return ChatMessageChunk(content=content, role=role)  # type: ignore[arg-type]
    return default_class(content=content)  # type: ignore[call-arg]


def _lc_tool_call_to_openai_tool_call(tool_call: ToolCall) -> dict:
    return {
        "type": "function",
        "id": tool_call["id"],
        "function": {
            "name": tool_call["name"],
            "arguments": json.dumps(tool_call["args"]),
        },
    }


def _convert_message_to_dict(message: BaseMessage) -> dict:
    message_dict: dict[str, Any] = {"content": message.content}
    if isinstance(message, ChatMessage):
        message_dict["role"] = message.role
    elif isinstance(message, HumanMessage):
        message_dict["role"] = "user"
    elif isinstance(message, AIMessage):
        message_dict["role"] = "assistant"
        if "function_call" in message.additional_kwargs:
            message_dict["function_call"] = message.additional_kwargs["function_call"]
        if message.tool_calls:
            message_dict["tool_calls"] = [
                _lc_tool_call_to_openai_tool_call(tc) for tc in message.tool_calls
            ]
        elif "tool_calls" in message.additional_kwargs:
            message_dict["tool_calls"] = message.additional_kwargs["tool_calls"]
    elif isinstance(message, SystemMessage):
        message_dict["role"] = "system"
    elif isinstance(message, FunctionMessage):
        message_dict["role"] = "function"
        message_dict["name"] = message.name
    elif isinstance(message, ToolMessage):
        message_dict["role"] = "tool"
        message_dict["tool_call_id"] = message.tool_call_id
    else:
        raise ValueError(f"Got unknown type {message}")
    if "name" in message.additional_kwargs:
        message_dict["name"] = message.additional_kwargs["name"]
    return message_dict


_OPENAI_MODELS = [
    "o1-mini",
    "o1-preview",
    "gpt-4o-mini",
    "gpt-4o-mini-2024-07-18",
    "gpt-4o",
    "gpt-4o-2024-08-06",
    "gpt-4o-2024-05-13",
    "gpt-4-turbo",
    "gpt-4-turbo-preview",
    "gpt-4-0125-preview",
    "gpt-4-1106-preview",
    "gpt-3.5-turbo-1106",
    "gpt-3.5-turbo",
    "gpt-3.5-turbo-0301",
    "gpt-3.5-turbo-0613",
    "gpt-3.5-turbo-16k",
    "gpt-3.5-turbo-16k-0613",
    "gpt-4",
    "gpt-4-0314",
    "gpt-4-0613",
    "gpt-4-32k",
    "gpt-4-32k-0314",
    "gpt-4-32k-0613",
    "gpt-4.1",
    "gpt-4.1-mini",
    "gpt-4.1-nano",
    "o3",
    "o3-mini",
    "o3-mini-high",
    "o4-mini",
    "o4-mini-high",
    "o1",
    "o1-mini",
    "o1-pro",
]


class ChatLiteLLM(BaseChatModel):
    """Chat model that uses the LiteLLM API."""

    client: Any = None  #: :meta private:
    model: str = "gpt-3.5-turbo"
    model_name: str | None = None
    """Model name to use."""
    openai_api_key: str | None = None
    azure_api_key: str | None = None
    anthropic_api_key: str | None = None
    replicate_api_key: str | None = None
    cohere_api_key: str | None = None
    openrouter_api_key: str | None = None
    api_key: str | None = None
    streaming: bool = False
    api_base: str | None = None
    organization: str | None = None
    custom_llm_provider: str | None = None
    request_timeout: float | tuple[float, float] | None = None
    temperature: float | None = None
    """Run inference with this temperature. Must be in the closed
       interval [0.0, 1.0]."""
    model_kwargs: dict[str, Any] = Field(default_factory=dict)
    """Holds any model parameters valid for API call not explicitly specified."""
    top_p: float | None = None
    """Decode using nucleus sampling: consider the smallest set of tokens whose
       probability sum is at least top_p. Must be in the closed interval [0.0, 1.0]."""
    top_k: int | None = None
    """Decode using top-k sampling: consider the set of top_k most probable tokens.
       Must be positive."""
    n: int | None = None
    """Number of chat completions to generate for each prompt. Note that the API may
       not return the full n completions if duplicates are generated."""
    max_tokens: int | None = None

    max_retries: int = 1

    @property
    def _default_params(self) -> dict[str, Any]:
        """Get the default parameters for calling OpenAI API."""
        set_model_value = self.model
        if self.model_name is not None:
            set_model_value = self.model_name
        return {
            "model": set_model_value,
            "force_timeout": self.request_timeout,
            "max_tokens": self.max_tokens,
            "stream": self.streaming,
            "n": self.n,
            "temperature": self.temperature,
            "custom_llm_provider": self.custom_llm_provider,
            **self.model_kwargs,
        }

    @property
    def _client_params(self) -> dict[str, Any]:
        """Get the parameters used for the openai client."""
        set_model_value = self.model
        if self.model_name is not None:
            set_model_value = self.model_name
        self.client.api_base = self.api_base
        self.client.api_key = self.api_key
        for named_api_key in [
            "openai_api_key",
            "azure_api_key",
            "anthropic_api_key",
            "replicate_api_key",
            "cohere_api_key",
            "openrouter_api_key",
        ]:
            if api_key_value := getattr(self, named_api_key):
                setattr(
                    self.client,
                    named_api_key.replace("_api_key", "_key"),
                    api_key_value,
                )
        self.client.organization = self.organization
        creds: dict[str, Any] = {
            "model": set_model_value,
            "force_timeout": self.request_timeout,
            "api_base": self.api_base,
        }
        return {**self._default_params, **creds}

    def completion_with_retry(
        self, run_manager: CallbackManagerForLLMRun | None = None, **kwargs: Any
    ) -> Any:
        """Use tenacity to retry the completion call."""
        retry_decorator = _create_retry_decorator(self, run_manager=run_manager)

        @retry_decorator
        def _completion_with_retry(**kwargs: Any) -> Any:
            return self.client.completion(**kwargs)

        return _completion_with_retry(**kwargs)

    @pre_init
    def validate_environment(cls, values: dict) -> dict:
        """Validate api key, python package exists, temperature, top_p, and top_k."""
        try:
            import litellm
        except ImportError:
            raise ChatLiteLLMException(
                "Could not import litellm python package. "
                "Please install it with `pip install litellm`"
            )

        values["openai_api_key"] = get_from_dict_or_env(
            values, "openai_api_key", "OPENAI_API_KEY", default=""
        )
        values["azure_api_key"] = get_from_dict_or_env(
            values, "azure_api_key", "AZURE_API_KEY", default=""
        )
        values["anthropic_api_key"] = get_from_dict_or_env(
            values, "anthropic_api_key", "ANTHROPIC_API_KEY", default=""
        )
        values["replicate_api_key"] = get_from_dict_or_env(
            values, "replicate_api_key", "REPLICATE_API_KEY", default=""
        )
        values["openrouter_api_key"] = get_from_dict_or_env(
            values, "openrouter_api_key", "OPENROUTER_API_KEY", default=""
        )
        values["cohere_api_key"] = get_from_dict_or_env(
            values, "cohere_api_key", "COHERE_API_KEY", default=""
        )
        values["huggingface_api_key"] = get_from_dict_or_env(
            values, "huggingface_api_key", "HUGGINGFACE_API_KEY", default=""
        )
        values["together_ai_api_key"] = get_from_dict_or_env(
            values, "together_ai_api_key", "TOGETHERAI_API_KEY", default=""
        )
        values["client"] = litellm

        if values["temperature"] is not None and not 0 <= values["temperature"] <= 1:
            raise ValueError("temperature must be in the range [0.0, 1.0]")

        if values["top_p"] is not None and not 0 <= values["top_p"] <= 1:
            raise ValueError("top_p must be in the range [0.0, 1.0]")

        if values["top_k"] is not None and values["top_k"] <= 0:
            raise ValueError("top_k must be positive")

        return values

    def _generate(
        self,
        messages: list[BaseMessage],
        stop: list[str] | None = None,
        run_manager: CallbackManagerForLLMRun | None = None,
        stream: bool | None = None,
        **kwargs: Any,
    ) -> ChatResult:
        should_stream = stream if stream is not None else self.streaming
        if should_stream:
            stream_iter = self._stream(
                messages, stop=stop, run_manager=run_manager, **kwargs
            )
            return generate_from_stream(stream_iter)

        message_dicts, params = self._create_message_dicts(messages, stop)
        params = {**params, **kwargs}
        response = self.completion_with_retry(
            messages=message_dicts, run_manager=run_manager, **params
        )
        return self._create_chat_result(response)

    def _create_chat_result(self, response: Mapping[str, Any]) -> ChatResult:
        generations = []
        token_usage = response.get("usage", {})
        for res in response["choices"]:
            message = _convert_dict_to_message(res["message"])
            if isinstance(message, AIMessage):
                message.response_metadata = {
                    "model_name": self.model_name or self.model
                }
                message.usage_metadata = _create_usage_metadata(token_usage)
            gen = ChatGeneration(
                message=message,
                generation_info=dict(finish_reason=res.get("finish_reason")),
            )
            generations.append(gen)
        set_model_value = self.model
        if self.model_name is not None:
            set_model_value = self.model_name
        llm_output = {"token_usage": token_usage, "model": set_model_value}
        return ChatResult(generations=generations, llm_output=llm_output)

    def _create_message_dicts(
        self, messages: list[BaseMessage], stop: list[str] | None
    ) -> tuple[list[dict[str, Any]], dict[str, Any]]:
        params = self._client_params
        if stop is not None:
            if "stop" in params:
                raise ValueError("`stop` found in both the input and default params.")
            params["stop"] = stop
        message_dicts = [_convert_message_to_dict(m) for m in messages]
        return message_dicts, params

    def _stream(
        self,
        messages: list[BaseMessage],
        stop: list[str] | None = None,
        run_manager: CallbackManagerForLLMRun | None = None,
        **kwargs: Any,
    ) -> Iterator[ChatGenerationChunk]:
        message_dicts, params = self._create_message_dicts(messages, stop)
        params = {**params, **kwargs, "stream": True}

        default_chunk_class = AIMessageChunk
        for chunk in self.completion_with_retry(
            messages=message_dicts, run_manager=run_manager, **params
        ):
            if not isinstance(chunk, dict):
                chunk = chunk.model_dump()
            if len(chunk["choices"]) == 0:
                continue
            delta = chunk["choices"][0]["delta"]
            chunk = _convert_delta_to_message_chunk(delta, default_chunk_class)
            default_chunk_class = chunk.__class__
            cg_chunk = ChatGenerationChunk(message=chunk)
            if run_manager:
                run_manager.on_llm_new_token(chunk.content, chunk=cg_chunk)
            yield cg_chunk

    async def _astream(
        self,
        messages: list[BaseMessage],
        stop: list[str] | None = None,
        run_manager: AsyncCallbackManagerForLLMRun | None = None,
        **kwargs: Any,
    ) -> AsyncIterator[ChatGenerationChunk]:
        message_dicts, params = self._create_message_dicts(messages, stop)
        params = {**params, **kwargs, "stream": True}

        default_chunk_class = AIMessageChunk
        async for chunk in await acompletion_with_retry(
            self, messages=message_dicts, run_manager=run_manager, **params
        ):
            if not isinstance(chunk, dict):
                chunk = chunk.model_dump()
            if len(chunk["choices"]) == 0:
                continue
            delta = chunk["choices"][0]["delta"]
            chunk = _convert_delta_to_message_chunk(delta, default_chunk_class)
            default_chunk_class = chunk.__class__
            cg_chunk = ChatGenerationChunk(message=chunk)
            if run_manager:
                await run_manager.on_llm_new_token(chunk.content, chunk=cg_chunk)
            yield cg_chunk

    async def _agenerate(
        self,
        messages: list[BaseMessage],
        stop: list[str] | None = None,
        run_manager: AsyncCallbackManagerForLLMRun | None = None,
        stream: bool | None = None,
        **kwargs: Any,
    ) -> ChatResult:
        should_stream = stream if stream is not None else self.streaming
        if should_stream:
            stream_iter = self._astream(
                messages=messages, stop=stop, run_manager=run_manager, **kwargs
            )
            return await agenerate_from_stream(stream_iter)

        message_dicts, params = self._create_message_dicts(messages, stop)
        params = {**params, **kwargs}
        response = await acompletion_with_retry(
            self, messages=message_dicts, run_manager=run_manager, **params
        )
        return self._create_chat_result(response)

    def bind_tools(
        self,
        tools: Sequence[dict[str, Any] | type[BaseModel] | Callable | BaseTool],
        tool_choice: dict
        | str
        | Literal["auto", "none", "required", "any"]
        | bool
        | None = None,
        **kwargs: Any,
    ) -> Runnable[LanguageModelInput, BaseMessage]:
        """Bind tool-like objects to this chat model.

        LiteLLM expects tools argument in OpenAI format.

        Args:
            tools: A list of tool definitions to bind to this chat model.
                Can be  a dictionary, pydantic model, callable, or BaseTool. Pydantic
                models, callables, and BaseTools will be automatically converted to
                their schema dictionary representation.
            tool_choice: Which tool to require the model to call. Options are:
                - str of the form ``"<<tool_name>>"``: calls <<tool_name>> tool.
                - ``"auto"``:
                    automatically selects a tool (including no tool).
                - ``"none"``:
                    does not call a tool.
                - ``"any"`` or ``"required"`` or ``True``:
                    forces least one tool to be called.
                - dict of the form:
                ``{"type": "function", "function": {"name": <<tool_name>>}}``
                - ``False`` or ``None``: no effect
            **kwargs: Any additional parameters to pass to the
                :class:`~langchain.runnable.Runnable` constructor.

        """
        formatted_tools = [convert_to_openai_tool(tool) for tool in tools]

        # In case of openai if tool_choice is `any` or if bool has been provided we
        # change it to `required` as that is supported by openai.
        if (
            (self.model is not None and "azure" in self.model)
            or (self.model_name is not None and "azure" in self.model_name)
            or (self.model is not None and self.model in _OPENAI_MODELS)
            or (self.model_name is not None and self.model_name in _OPENAI_MODELS)
        ) and (tool_choice == "any" or isinstance(tool_choice, bool)):
            tool_choice = "required"
        # If tool_choice is bool apart from openai we make it `any`
        elif isinstance(tool_choice, bool):
            tool_choice = "any"
        elif isinstance(tool_choice, dict):
            tool_names = [
                formatted_tool["function"]["name"] for formatted_tool in formatted_tools
            ]
            if not any(
                tool_name == tool_choice["function"]["name"] for tool_name in tool_names
            ):
                raise ValueError(
                    f"Tool choice {tool_choice} was specified, but the only "
                    f"provided tools were {tool_names}."
                )
        return super().bind(tools=formatted_tools, tool_choice=tool_choice, **kwargs)

    @property
    def _identifying_params(self) -> dict[str, Any]:
        """Get the identifying parameters."""
        set_model_value = self.model
        if self.model_name is not None:
            set_model_value = self.model_name
        return {
            "model": set_model_value,
            "temperature": self.temperature,
            "top_p": self.top_p,
            "top_k": self.top_k,
            "n": self.n,
        }

    @property
    def _llm_type(self) -> str:
        return "litellm-chat"


def _create_usage_metadata(token_usage: Mapping[str, Any]) -> UsageMetadata:
    input_tokens = token_usage.get("prompt_tokens", 0)
    output_tokens = token_usage.get("completion_tokens", 0)
    return UsageMetadata(
        input_tokens=input_tokens,
        output_tokens=output_tokens,
        total_tokens=input_tokens + output_tokens,
    )



from importlib.metadata import PackageNotFoundError, version

from .config import AgentConfig, AgentFramework
from .frameworks.any_agent import AgentRunError, AnyAgent
from .tracing.agent_trace import AgentTrace

try:
    __version__ = version("any-agent")
except PackageNotFoundError:
    # In the case of local development
    # i.e., running directly from the source directory without package being installed
    __version__ = "0.0.0-dev"

__all__ = [
    "AgentConfig",
    "AgentFramework",
    "AgentRunError",
    "AgentTrace",
    "AnyAgent",
    "__version__",
]



# adapted from https://github.com/google/a2a-python/blob/main/examples/helloworld/test_client.py

import re
from collections.abc import Callable, Coroutine
from contextlib import suppress
from typing import TYPE_CHECKING, Any, Optional
from uuid import uuid4

from any_agent.utils.asyncio_sync import run_async_in_sync

if TYPE_CHECKING:
    from a2a.types import AgentCard

a2a_tool_available = False
with suppress(ImportError):
    import httpx
    from a2a.client import A2ACardResolver, A2AClient
    from a2a.types import (
        JSONRPCErrorResponse,
        Message,
        MessageSendParams,
        Part,
        Role,
        SendMessageRequest,
        SendMessageSuccessResponse,
        Task,
        TextPart,
    )

    a2a_tool_available = True


async def a2a_tool_async(
    url: str, toolname: Optional[str] = None, http_kwargs: dict[str, Any] | None = None
) -> Callable[[str, Optional[str], Optional[str]], Coroutine[Any, Any, dict[str, Any]]]:
    """Perform a query using A2A to another agent.

    Args:
        url (str): The url in which the A2A agent is located.
        toolname (str): The name for the created tool. Defaults to `call_{agent name in card}`.
            Leading and trailing whitespace are removed. Whitespace in the middle is replaced by `_`.
        http_kwargs (dict): Additional kwargs to pass to the httpx client.

    Returns:
        An async `Callable` that takes a query and returns the agent response.

    """
    if not a2a_tool_available:
        msg = "You need to `pip install 'any-agent[a2a]'` to use this tool"
        raise ImportError(msg)

    if http_kwargs is None:
        http_kwargs = {}

    # Default timeout in httpx is 5 seconds. For an agent response, the default should be more lenient.
    if "timeout" not in http_kwargs:
        http_kwargs["timeout"] = 30.0

    async with httpx.AsyncClient(
        follow_redirects=True, **http_kwargs
    ) as resolver_client:
        a2a_agent_card: AgentCard = await (
            A2ACardResolver(httpx_client=resolver_client, base_url=url)
        ).get_agent_card()

    # NOTE: Use Optional[T] instead of T | None syntax throughout this module.
    # Google ADK's _parse_schema_from_parameter function has compatibility
    # with the traditional Optional[T] syntax for automatic function calling.
    # Using T | None syntax causes"Failed to parse the parameter ... for automatic function calling"
    async def _send_query(
        query: str, task_id: Optional[str] = None, context_id: Optional[str] = None
    ) -> dict[str, Any]:
        async with httpx.AsyncClient(follow_redirects=True) as query_client:
            client = A2AClient(httpx_client=query_client, agent_card=a2a_agent_card)
            send_message_payload = SendMessageRequest(
                id=str(uuid4()),
                params=MessageSendParams(
                    message=Message(
                        role=Role.user,
                        parts=[Part(root=TextPart(text=query))],
                        # the id is not currently tracked
                        messageId=str(uuid4().hex),
                        taskId=task_id,
                        contextId=context_id,
                    )
                ),
            )
            # TODO check how to capture exceptions and pass them on to the enclosing framework
            response = await client.send_message(
                send_message_payload, http_kwargs=http_kwargs
            )

            if not response.root:
                msg = (
                    "The A2A agent did not return a root. Are you using an A2A agent not managed by any-agent? "
                    "Please file an issue at https://github.com/mozilla-ai/any-agent/issues so we can help."
                )
                raise ValueError(msg)

            if isinstance(response.root, JSONRPCErrorResponse):
                response_dict = {
                    "error": response.root.error.message,
                    "code": response.root.error.code,
                    "data": response.root.error.data,
                }
            elif isinstance(response.root, SendMessageSuccessResponse):
                # Task
                if isinstance(response.root.result, Task):
                    task = response.root.result
                    response_dict = {
                        "timestamp": task.status.timestamp,
                        "status": task.status.state,
                    }
                    if task.status.message:
                        response_dict["task_id"] = task.status.message.taskId
                        response_dict["context_id"] = task.status.message.contextId
                        response_dict["message"] = {
                            " ".join(
                                [
                                    part.root.text
                                    for part in task.status.message.parts
                                    if isinstance(part.root, TextPart)
                                ]
                            )
                        }
                # Message
                else:
                    response_dict = {
                        "message": {
                            " ".join(
                                [
                                    part.root.text
                                    for part in response.root.result.parts
                                    if isinstance(part.root, TextPart)
                                ]
                            )
                        },
                        "task_id": response.root.result.taskId,
                    }
            else:
                msg = (
                    "The A2A agent did not return a error or a result. Are you using an A2A agent not managed by any-agent? "
                    "Please file an issue at https://github.com/mozilla-ai/any-agent/issues so we can help."
                )
                raise ValueError(msg)

            return response_dict

    new_name = toolname or a2a_agent_card.name
    new_name = re.sub(r"\s+", "_", new_name.strip())
    _send_query.__name__ = f"call_{new_name}"
    _send_query.__doc__ = f"""{a2a_agent_card.description}
        Send a query to the A2A hosted agent named {a2a_agent_card.name}.

        Agent description: {a2a_agent_card.description}

        Args:
            query (str): The query to send to the agent.
            task_id (str, optional): Task ID for continuing an incomplete task. Use the same
                task_id from a previous response with TaskState.input_required to resume the task. If you want to start a new task, you should not provide a task id.
            context_id (str, optional): Context ID for conversation continuity. Provides the
                agent with conversation history. Omit to start a fresh conversation. If you want to start a new conversation, you should not provide a context id.

        Returns:
            dict: Response from the A2A agent containing:
                - For successful responses: task_id, context_id, timestamp, status, and message
                - For errors: error message, code, and data

        Note:
            If TaskState is terminal (completed/failed), do not reuse the same task_id.
    """
    return _send_query


def a2a_tool(
    url: str, toolname: Optional[str] = None, http_kwargs: dict[str, Any] | None = None
) -> Callable[[str, Optional[str], Optional[str]], str]:
    """Perform a query using A2A to another agent (synchronous version).

    Args:
        url (str): The url in which the A2A agent is located.
        toolname (str): The name for the created tool. Defaults to `call_{agent name in card}`.
            Leading and trailing whitespace are removed. Whitespace in the middle is replaced by `_`.
        http_kwargs (dict): Additional kwargs to pass to the httpx client.

    Returns:
        A sync `Callable` that takes a query and returns the agent response.

    """
    if not a2a_tool_available:
        msg = "You need to `pip install 'any-agent[a2a]'` to use this tool"
        raise ImportError(msg)

    # Fetch the async tool upfront to get proper name and documentation (otherwise the tool doesn't have the right name and documentation)
    async_tool = run_async_in_sync(a2a_tool_async(url, toolname, http_kwargs))

    def sync_wrapper(
        query: str, task_id: Optional[str] = None, context_id: Optional[str] = None
    ) -> Any:
        """Execute the A2A tool query synchronously."""
        return run_async_in_sync(async_tool(query, task_id, context_id))

    # Copy essential metadata from the async tool
    sync_wrapper.__name__ = async_tool.__name__
    sync_wrapper.__doc__ = async_tool.__doc__

    return sync_wrapper



from collections.abc import Callable
from typing import Any

from pydantic import BaseModel

from any_agent import AgentConfig, AnyAgent
from any_agent.config import AgentFramework
from any_agent.evaluation.schemas import EvaluationOutput
from any_agent.evaluation.tools import TraceTools
from any_agent.tracing.agent_trace import AgentTrace
from any_agent.utils.asyncio_sync import run_async_in_sync

AGENT_INSTRUCTIONS = """You are a helpful assistant that will be used to evaluate the correctness of an agent trace.
Given a specific question regarding the quality of something about the agent, \
you may utilize tools as needed in order to check if the trace satisfies the question.

Answer with:
1. "passed": true or false (true if the trace satisfies the question, false otherwise)
2. "reasoning": Brief explanation for your decision (2-3 sentences max)

Your output must match the following JSON schema:
{response_schema}"""


class AgentJudge:
    """An agent that evaluates the correctness of another agent's trace."""

    def __init__(
        self,
        model_id: str,
        framework: AgentFramework = AgentFramework.TINYAGENT,
        output_type: type[BaseModel] = EvaluationOutput,
        model_args: dict[str, Any] | None = None,
    ):
        self.model_id = model_id
        self.framework = framework
        self.model_args = model_args
        self.output_type = output_type

    def run(
        self,
        trace: AgentTrace,
        question: str,
        additional_tools: list[Callable[[], Any]] | None = None,
    ) -> AgentTrace:
        """Run the agent judge.

        Args:
            trace: The agent trace to evaluate
            question: The question to ask the agent
            additional_tools: Additional tools to use for the agent

        Returns:
            The trace of the evaluation run.
            You can access the evaluation result in the `final_output`
            property.

        """
        if additional_tools is None:
            additional_tools = []
        return run_async_in_sync(self.run_async(trace, question, additional_tools))

    async def run_async(
        self,
        trace: AgentTrace,
        question: str,
        additional_tools: list[Callable[[], Any]] | None = None,
    ) -> AgentTrace:
        """Run the agent judge asynchronously.

        Args:
            trace: The agent trace to evaluate
            question: The question to ask the agent
            additional_tools: Additional tools to use for the agent
        Returns:
            The trace of the evaluation run.
            You can access the evaluation result in the `final_output`
            property.

        """
        if additional_tools is None:
            additional_tools = []
        tooling = TraceTools(trace)

        agent_config = AgentConfig(
            model_id=self.model_id,
            instructions=AGENT_INSTRUCTIONS.format(
                response_schema=self.output_type.model_json_schema()
            ),
            tools=tooling.get_all_tools() + additional_tools,
            output_type=self.output_type,
            model_args=self.model_args,
        )

        agent = await AnyAgent.create_async(
            self.framework,
            agent_config=agent_config,
        )
        agent_trace = await agent.run_async(question)
        if not isinstance(agent_trace.final_output, self.output_type):
            msg = f"Agent output is not an {self.output_type} instance."
            raise ValueError(msg)
        return agent_trace



from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any

from opentelemetry.trace import Status, StatusCode

from any_agent.callbacks.base import Callback
from any_agent.tracing.attributes import GenAI

if TYPE_CHECKING:
    from any_agent.callbacks.context import Context


class _SpanGeneration(Callback):
    def __init__(self) -> None:
        self.first_llm_calls: set[int] = set()

    def _serialize_for_attribute(self, data: Any) -> str:
        """Serialize data for OpenTelemetry attributes, handling various types safely."""
        if isinstance(data, str):
            return data
        try:
            return json.dumps(data, default=str, ensure_ascii=False)
        except (TypeError, ValueError):
            return str(data)

    def _determine_output_type(self, output: Any) -> str:
        """Determine output type based on the output content."""
        if isinstance(output, str):
            try:
                json.loads(output)
            except json.JSONDecodeError:
                return "text"
        return "json"

    def _set_llm_input(
        self, context: Context, model_id: str, input_messages: list[dict[str, str]]
    ) -> Context:
        tracer = context.tracer
        span = tracer.start_span(f"call_llm {model_id}")

        span.set_attributes(
            {
                GenAI.OPERATION_NAME: "call_llm",
                GenAI.REQUEST_MODEL: model_id,
            }
        )

        trace_id = span.get_span_context().trace_id
        if trace_id not in self.first_llm_calls:
            self.first_llm_calls.add(trace_id)
            serialized_messages = self._serialize_for_attribute(input_messages)
            span.set_attribute(GenAI.INPUT_MESSAGES, serialized_messages)

        context.current_span = span
        return context

    def _set_llm_output(
        self,
        context: Context,
        output: str | list[dict[str, str]],
        input_tokens: int,
        output_tokens: int,
    ) -> Context:
        span = context.current_span
        output_type = self._determine_output_type(output)
        output_attr = self._serialize_for_attribute(output)

        span.set_attributes(
            {
                GenAI.OUTPUT: output_attr,
                GenAI.OUTPUT_TYPE: output_type,
                GenAI.USAGE_INPUT_TOKENS: input_tokens,
                GenAI.USAGE_OUTPUT_TOKENS: output_tokens,
            }
        )

        span.set_status(StatusCode.OK)
        return context

    def _set_tool_input(
        self,
        context: Context,
        name: str,
        description: str | None = None,
        args: dict[str, Any] | None = None,
        call_id: str | None = None,
    ) -> Context:
        tracer = context.tracer
        span = tracer.start_span(f"execute_tool {name}")

        attributes = {
            GenAI.OPERATION_NAME: "execute_tool",
            GenAI.TOOL_NAME: name,
        }

        if description is not None:
            attributes[GenAI.TOOL_DESCRIPTION] = description
        if args is not None:
            attributes[GenAI.TOOL_ARGS] = self._serialize_for_attribute(args)
        if call_id is not None:
            attributes["gen_ai.tool.call.id"] = call_id

        span.set_attributes(attributes)
        context.current_span = span
        return context

    def _determine_tool_status(
        self, tool_output: str, output_type: str
    ) -> Status | StatusCode:
        """Determine the status based on tool output content and type."""
        if output_type == "text" and "Error calling tool:" in tool_output:
            return Status(StatusCode.ERROR, description=tool_output)
        return StatusCode.OK

    def _set_tool_output(self, context: Context, tool_output: Any) -> Context:
        span = context.current_span

        if tool_output is None:
            tool_output = "{}"

        output_type = self._determine_output_type(tool_output)
        output_attr = self._serialize_for_attribute(tool_output)
        status = self._determine_tool_status(output_attr, output_type)

        span.set_attributes({GenAI.OUTPUT: output_attr, GenAI.OUTPUT_TYPE: output_type})
        span.set_status(status)
        return context
